package biggests.clustering;

import smadeira.biclustering.DiscretizedExpressionMatrix;
import smadeira.biclustering.AbstractExpressionMatrix;
import smadeira.biclustering.ExpressionMatrix;
import smadeira.biclustering.IMatrix;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * <p>Title: Abstract Metric Clustering</p>
 *
 * <p>Description: Defines an abstract class for clustering
 *                 methods based on distance metrics.</p>
 *
 * <p>Copyright:   Copyright (C) 2008  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public abstract class AbstractMetricClustering {
    /**
     * Clustering on genes (rows) dimension (clustering dimension).
     */
    public static final boolean GENES_DIMENSION = false;

    /**
     * Clustering on conditions (columns) dimension (clustering dimension).
     */
    public static final boolean CONDITIONS_DIMENSION = true;

    /**
     * No clustering (distance metric).
     */
    public static final char NO_CLUSTERING = '\0';

    /**
     * Uncentered correlation (distance metric).
     */
    public static final char UNCENTERED_CORRELATION = 'u';

    /**
     * Pearson's correlation (distance metric).
     */
    public static final char PEARSON_CORRELATION = 'c';

    /**
     * Uncentered absolute correlation (distance metric).
     */
    public static final char UNCENTERED_ABS_CORRELATION = 'x';

    /**
     * Pearson's absolute correlation (distance metric).
     */
    public static final char PEARSON_ABS_CORRELATION = 'a';

    /**
     * Spearman's rank correlation (distance metric).
     */
    public static final char SPEARMAN_RANK_CORRELATION = 's';

    /**
     * Kendall's tau correlation (distance metric).
     */
    public static final char KENDALL_TAU_CORRELATION = 'k';

    /**
     * Euclidean distance (distance metric).
     */
    public static final char EUCLIDEAN_DISTANCE = 'e';

    /**
     * Cityblock distance (distance metric).
     */
    public static final char CITYBLOCK_DISTANCE = 'b';

    /**
     * The number of genes in the experiment.
     */
    protected int numberOfGenes;

    /**
     * The number of conditions (microarrays) in the experiment.
     */
    protected int numberOfConditions;

    /**
     * The weights of the genes.
     */
    protected double[] genesWeights;

    /**
     * The weights of the conditions.
     */
    protected double[] conditionsWeights;

    /**
     * The order of the genes in the data file.
     */
    protected double[] genesOrder;

    /**
     * The order of the conditions in the data file.
     */
    protected double[] conditionsOrder;

    /**
     * The indexes of the genes set by clustering method for file output.
     */
    protected int[] genesIndexes;

    /**
     * The indexes of the conditions set by clustering method for file output.
     */
    protected int[] conditionsIndexes;

    /**
     * The name of the unique ID used to identify genes in the data file.
     */
    protected String uniqueID;

    /**
     * Genes' unique IDs.
     */
    protected String[] genesUniqueIDs;

    /**
     * Genes' names.
     */
    protected String[] genesNames;

    /**
     * Conditions' names.
     */
    protected String[] conditionsNames;

    /**
     * Expression data. Contains the expression values measured
     * for the genes in the experimental conditions.
     */
    protected double[][] expressionData;

    /**
     * Missing values mask.<br/>
     * <code>missingsMask[i][j] == 1</code> -> valid expression value in
     * <code>expressionData[i][j]</code>
     * <code>missingsMask[i][j] == 0</code> -> missing expression value in
     * <code>expressionData[i][j]</code> (this means that the value in
     * <code>expressionData[i][j]</code> should not be used for calculations).
     * <br/>
     * Mask is composed by integer values, not boolean, because these
     * values are sometimes used in calculation.
     */
    protected int[][] missingsMask;

    /**
     * Creates a new <code>AbstractClustering</code> object to
     * perform clustering over the set of gene expression
     * data contained in <code>expressionMatrix</code>.
     *
     * @param expressionMatrix <code>IMatrix</code>
     */
    public AbstractMetricClustering(IMatrix expressionMatrix) {
        AbstractExpressionMatrix matrix;
        if (expressionMatrix instanceof DiscretizedExpressionMatrix) {
            matrix = ((DiscretizedExpressionMatrix) expressionMatrix).
                     getOriginalExpressionMatrix();
        } else {
            matrix = (AbstractExpressionMatrix) expressionMatrix;
        }
        numberOfGenes = matrix.getNumberOfGenes();
        numberOfConditions = matrix.getNumberOfConditions();
        genesWeights = new double[numberOfGenes];
        conditionsWeights = new double[numberOfConditions];
        genesOrder = new double[numberOfGenes];
        conditionsOrder = new double[numberOfConditions];
        genesIndexes = new int[numberOfGenes];
        conditionsIndexes = new int[numberOfConditions];
        uniqueID = "GENES";
        genesUniqueIDs = new String[numberOfGenes];
        genesNames = matrix.getGenesNames();
        conditionsNames = matrix.getConditionsNames();

        expressionData = new double[numberOfGenes][numberOfConditions];
        float[][] x = matrix.getGeneExpressionMatrix();
        for (int i = 0; i < expressionData.length; i++) {
            for (int j = 0; j < expressionData[0].length; j++) {
                expressionData[i][j] = x[i][j];
            }
        }

        missingsMask = new int[numberOfGenes][numberOfConditions];
        if (matrix instanceof ExpressionMatrix &&
            ((ExpressionMatrix) matrix).hasMissingValues()) {
            float value = ((ExpressionMatrix) matrix).getMissingValue();
            for (int i = 0; i < x.length; i++) {
                for (int j = 0; j < x[0].length; j++) {
                    if (x[i][j] == value) {
                        missingsMask[i][j] = 0;
                    } else {
                        missingsMask[i][j] = 1;
                    }
                }
            }
        } else {
            for (int i = 0; i < missingsMask.length; i++) {
                for (int j = 0; j < missingsMask[0].length; j++) {
                    missingsMask[i][j] = 1;
                }
            }
        }

        for (int i = 0; i < genesWeights.length; i++) {
            genesWeights[i] = 1;
            genesOrder[i] = i;
            genesIndexes[i] = i;
            genesUniqueIDs[i] = String.valueOf("ID" + i);
        }
        for (int i = 0; i < conditionsWeights.length; i++) {
            conditionsWeights[i] = 1;
            conditionsOrder[i] = i;
            conditionsIndexes[i] = i;
        }
    }

    /**
     * Creates a new <code>AbstractMetricClustering</code> object to
     * perform clustering over the given set of
     * expression data.
     *
     * @param numberOfGenes <code>int</code> the number of genes in the
     *        experiment
     * @param numberOfConditions <code>int</code> the number of conditions
     *        in the experiment
     * @param genesWeights <code>double[]</code> the set of genes' weights
     * @param conditionsWeights <code>double[]</code> the set of conditions'
     *        weights
     * @param genesOrder <code>double[]</code> the order of the genes
     * @param conditionsOrder <code>double[]</code> the order of the conditions
     * @param genesIndexes <code>int[]</code> the set of genes' indexes
     * @param conditionsIndexes <code>int[]</code> the set of conditions'
     *        indexes
     * @param uniqueID <code>String</code> the unique ID type name
     * @param genesUniqueIDs <code>String[]</code> the set of genes' unique IDs
     * @param genesNames <code>String[]</code> the set of genes' names
     * @param conditionsNames <code>String[]</code> the set of conditions' names
     * @param expressionData <code>double[][]</code> the set of expression
     *        values measured for the genes in the experimental conditions
     * @param missingsMask <code>int[][]</code> the mask which states if each
     *        element contained in <code>expressionData</code> is valid (1) or
     *        missing (0)
     * @throws Exception - if array and/or matrices dimensions do not match
     *        the <code>numberOfGenes</code> and/or
     *        <code>numberOfConditions</code>
     */
    public AbstractMetricClustering(int numberOfGenes,
                                  int numberOfConditions,
                                  double[] genesWeights,
                                  double[] conditionsWeights,
                                  double[] genesOrder,
                                  double[] conditionsOrder,
                                  int[] genesIndexes,
                                  int[] conditionsIndexes,
                                  String uniqueID,
                                  String[] genesUniqueIDs,
                                  String[] genesNames,
                                  String[] conditionsNames,
                                  double[][] expressionData,
                                  int[][] missingsMask) throws Exception {
        if (genesWeights.length != numberOfGenes ||
            genesOrder.length != numberOfGenes ||
            genesIndexes.length != numberOfGenes ||
            genesUniqueIDs.length != numberOfGenes ||
            genesNames.length != numberOfGenes ||
            expressionData.length != numberOfGenes ||
            missingsMask.length != numberOfGenes) {
            throw new Exception("Number of genes does not match in the " +
                                "given information. The length of " +
                                "genesWeights, genesOrder, genesIndexes, " +
                                "genesUniqueIDs and genesNames arrays must " +
                                "equal the numberOfGenes. Also the number " +
                                "of rows (first dimension) of expressionData " +
                                "and missingsMask matrices (bidimensional " +
                                "arrays) must equal the numberOfGenes.");
        }
        if (conditionsWeights.length != numberOfConditions ||
            conditionsOrder.length != numberOfConditions ||
            conditionsIndexes.length != numberOfConditions ||
            conditionsNames.length != numberOfConditions ||
            expressionData[0].length != numberOfConditions ||
            missingsMask[0].length != numberOfConditions) {
            throw new Exception("Number of genes does not match in the " +
                                "given information. The length of " +
                                "conditionsWeights, conditionsOrder, " +
                                "conditionsIndexes, and conditionsNames " +
                                "arrays must equal the numberOfGenes. " +
                                "Also the number of rows (first dimension)" +
                                " of expressionData and missingsMask " +
                                "matrices (bidimensional " +
                                "arrays) must equal the numberOfGenes.");
        }

        this.numberOfGenes = numberOfGenes;
        this.numberOfConditions = numberOfConditions;
        this.genesWeights = genesWeights;
        this.conditionsWeights = conditionsWeights;
        this.genesOrder = genesOrder;
        this.conditionsOrder = conditionsOrder;
        this.genesIndexes = genesIndexes;
        this.conditionsIndexes = conditionsIndexes;
        this.uniqueID = uniqueID;
        this.genesUniqueIDs = genesUniqueIDs;
        this.genesNames = genesNames;
        this.conditionsNames = conditionsNames;
        this.expressionData = expressionData;
        this.missingsMask = missingsMask;
    }

    /**
     * Calculates the weighted Euclidean distance between two genes/rows or two
     * conditions/columns. To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double euclideanDistance(double[][] expressionData1,
                                           double[][] expressionData2,
                                           int[][] missingsMask1,
                                           int[][] missingsMask2,
                                           double[] elementsWeights,
                                           int index1,
                                           int index2,
                                           boolean dimension) throws Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);

        double result = 0.;
        double tweight = 0;
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION)
                /* Calculate the distance between two rows */ {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    double term = expressionData1[index1][i] -
                                  expressionData2[index2][i];
                    result += elementsWeights[i] * term * term;
                    tweight += elementsWeights[i];
                }
            }
        } else {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    double term = expressionData1[i][index1] -
                                  expressionData2[i][index2];
                    result += elementsWeights[i] * term * term;
                    tweight += elementsWeights[i];
                }
            }
        }
        if (tweight == 0) {
            return 0; /* usually due to empty clusters */
        }
        result /= tweight;
        return result;
    }

    /**
     * Calculates the weighted "City Block" distance between two
     * genes/rows or conditions/columns in a matrix. City Block distance
     * is defined as the absolute value of X1-X2 plus the absolute value
     * of Y1-Y2 plus..., which is equivalent to taking an "up and over"
     * path.<br/>
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double cityblockDistance(double[][] expressionData1,
                                           double[][] expressionData2,
                                           int[][] missingsMask1,
                                           int[][] missingsMask2,
                                           double[] elementsWeights,
                                           int index1,
                                           int index2,
                                           boolean dimension) throws Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);
        double result = 0.;
        double tweight = 0;
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION)
                /* Calculate the distance between two rows */ {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    double term = expressionData1[index1][i] -
                                  expressionData2[index2][i];
                    result = result + elementsWeights[i] * Math.abs(term);
                    tweight += elementsWeights[i];
                }
            }
        } else {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    double term = expressionData1[i][index1] -
                                  expressionData2[i][index2];
                    result = result + elementsWeights[i] * Math.abs(term);
                    tweight += elementsWeights[i];
                }
            }
        }
        if (tweight == 0) {
            return 0; /* usually due to empty clusters */
        }
        result /= tweight;
        return result;
    }

    /**
     * Calculates the weighted Pearson distance between two
     * genes/rows or conditions/columns in a matrix. We define
     * the Pearson distance as one minus the Pearson correlation.
     * This definition yields a semi-metric: d(a,b) >= 0, and
     * d(a,b) = 0 iff a = b. but the triangular inequality
     * d(a,b) + d(b,c) >= d(a,c) does not hold
     * (e.g., choose b = a + c).<br/>
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double pearsonsCorrelation(double[][] expressionData1,
                                             double[][] expressionData2,
                                             int[][] missingsMask1,
                                             int[][] missingsMask2,
                                             double elementsWeights[],
                                             int index1,
                                             int index2,
                                             boolean dimension) throws
            Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);

        double result = 0.;
        double sum1 = 0.;
        double sum2 = 0.;
        double denom1 = 0.;
        double denom2 = 0.;
        double tweight = 0.;
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION)
                /* Calculate the distance between two rows */ {
            int i;
            for (i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    double term1 = expressionData1[index1][i];
                    double term2 = expressionData2[index2][i];
                    double w = elementsWeights[i];
                    sum1 += w * term1;
                    sum2 += w * term2;
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    tweight += w;
                }
            }
        } else {
            int i;
            for (i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    double term1 = expressionData1[i][index1];
                    double term2 = expressionData2[i][index2];
                    double w = elementsWeights[i];
                    sum1 += w * term1;
                    sum2 += w * term2;
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    tweight += w;
                }
            }
        }
        if (tweight == 0) {
            return 0; /* usually due to empty clusters */
        }
        result -= sum1 * sum2 / tweight;
        denom1 -= sum1 * sum1 / tweight;
        denom2 -= sum2 * sum2 / tweight;
        if (denom1 <= 0) {
            return 1; /* include '<' to deal with roundoff errors */
        }
        if (denom2 <= 0) {
            return 1; /* include '<' to deal with roundoff errors */
        }
        result = result / Math.sqrt(denom1 * denom2);
        result = 1. - result;
        return result;
    }

    /**
     * Calculates the weighted Pearson distance between two
     * genes/rows or conditions/columns, using the absolute
     * value of the correlation. This definition yields a
     * semi-metric: d(a,b) >= 0, and d(a,b) = 0 iff a = b.
     * but the triangular inequality d(a,b) + d(b,c) >= d(a,c)
     * does not hold (e.g., choose b = a + c).<br/>
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double pearsonsAbsoluteCorrelation(double[][] expressionData1,
            double[][] expressionData2,
            int[][] missingsMask1,
            int[][] missingsMask2,
            double[] elementsWeights,
            int index1,
            int index2,
            boolean dimension) throws Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);

        double result = 0.;
        double sum1 = 0.;
        double sum2 = 0.;
        double denom1 = 0.;
        double denom2 = 0.;
        double tweight = 0.;
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION)
                /* Calculate the distance between two rows */ {
            int i;
            for (i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    double term1 = expressionData1[index1][i];
                    double term2 = expressionData2[index2][i];
                    double w = elementsWeights[i];
                    sum1 += w * term1;
                    sum2 += w * term2;
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    tweight += w;
                }
            }
        } else {
            int i;
            for (i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    double term1 = expressionData1[i][index1];
                    double term2 = expressionData2[i][index2];
                    double w = elementsWeights[i];
                    sum1 += w * term1;
                    sum2 += w * term2;
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    tweight += w;
                }
            }
        }
        if (tweight == 0) {
            return 0; /* usually due to empty clusters */
        }
        result -= sum1 * sum2 / tweight;
        denom1 -= sum1 * sum1 / tweight;
        denom2 -= sum2 * sum2 / tweight;
        if (denom1 <= 0) {
            return 1; /* include '<' to deal with roundoff errors */
        }
        if (denom2 <= 0) {
            return 1; /* include '<' to deal with roundoff errors */
        }
        result = Math.abs(result) / Math.sqrt(denom1 * denom2);
        result = 1. - result;
        return result;
    }

    /**
     * Calculates the weighted Pearson distance between two
     * genes/rows or conditions/columns, using the uncentered
     * version of the Pearson correlation. In the uncentered
     * Pearson correlation, a zero mean is used for both
     * vectors even if the actual mean is nonzero. This
     * definition yields a semi-metric: d(a,b) >= 0, and
     * d(a,b) = 0 iff a = b. but the triangular
     * inequality d(a,b) + d(b,c) >= d(a,c) does not hold
     * (e.g., choose b = a + c).<br/>
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double uncenteredCorrelation(double[][] expressionData1,
                                               double[][] expressionData2,
                                               int[][] missingsMask1,
                                               int[][] missingsMask2,
                                               double[] elementsWeights,
                                               int index1,
                                               int index2,
                                               boolean dimension) throws
            Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);

        double result = 0.;
        double denom1 = 0.;
        double denom2 = 0.;
        boolean flag = false;
        /* flag will remain zero if no nonzero combinations of mask1 and mask2 are
         * found.
         */
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION)
                /* Calculate the distance between two rows */ {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    double term1 = expressionData1[index1][i];
                    double term2 = expressionData2[index2][i];
                    double w = elementsWeights[i];
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    flag = true;
                }
            }
        } else {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    double term1 = expressionData1[i][index1];
                    double term2 = expressionData2[i][index2];
                    double w = elementsWeights[i];
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    flag = true;
                }
            }
        }
        if (!flag) {
            return 0.;
        }
        if (denom1 == 0.) {
            return 1.;
        }
        if (denom2 == 0.) {
            return 1.;
        }
        result = result / Math.sqrt(denom1 * denom2);
        result = 1. - result;
        return result;
    }

    /**
     * Calculates the weighted Pearson distance between two
     * genes/rows or conditions/columns, using the absolute
     * value of the uncentered version of the Pearson correlation.
     * In the uncentered Pearson correlation, a zero mean is used
     * for both vectors even if the actual mean is nonzero.
     * This definition yields a semi-metric: d(a,b) >= 0,
     * and d(a,b) = 0 iff a = b. but the triangular inequality
     * d(a,b) + d(b,c) >= d(a,c) does not hold (e.g., choose b = a + c).<br/>
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double uncenteredAbsoluteCorrelation(double[][]
            expressionData1,
            double[][] expressionData2,
            int[][] missingsMask1,
            int[][] missingsMask2,
            double[] elementsWeights,
            int index1,
            int index2,
            boolean dimension) throws Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);

        double result = 0.;
        double denom1 = 0.;
        double denom2 = 0.;
        boolean flag = false;
        /* flag will remain zero if no nonzero combinations of mask1 and mask2 are
         * found.
         */
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION)
                /* Calculate the distance between two rows */ {
            int i;
            for (i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    double term1 = expressionData1[index1][i];
                    double term2 = expressionData2[index2][i];
                    double w = elementsWeights[i];
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    flag = true;
                }
            }
        } else {
            int i;
            for (i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    double term1 = expressionData1[i][index1];
                    double term2 = expressionData2[i][index2];
                    double w = elementsWeights[i];
                    result += w * term1 * term2;
                    denom1 += w * term1 * term1;
                    denom2 += w * term2 * term2;
                    flag = true;
                }
            }
        }
        if (!flag) {
            return 0.;
        }
        if (denom1 == 0.) {
            return 1.;
        }
        if (denom2 == 0.) {
            return 1.;
        }
        result = Math.abs(result) / Math.sqrt(denom1 * denom2);
        result = 1. - result;
        return result;
    }

    /**
     * Calculates the Spearman distance between two genes/rows or
     * conditions/columns. The Spearman distance is defined as one
     * minus the Spearman rank correlation.<br/>
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double spearmansRankCorrelation(double[][] expressionData1,
                                                  double[][] expressionData2,
                                                  int[][] missingsMask1,
                                                  int[][] missingsMask2,
                                                  double[] elementsWeights,
                                                  int index1,
                                                  int index2,
                                                  boolean dimension) throws
            Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);

        int m = 0;
        double[] ranks1;
        double[] ranks2;
        double result = 0.;
        double denom1 = 0.;
        double denom2 = 0.;
        double avgrank;
        ArrayList<Double> tdata1 = new ArrayList<Double>();
        ArrayList<Double> tdata2 = new ArrayList<Double>();
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION) {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    tdata1.add(expressionData1[index1][i]);
                    tdata2.add(expressionData2[index2][i]);
                    m++;
                }
            }
        } else {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    tdata1.add(expressionData1[i][index1]);
                    tdata2.add(expressionData2[i][index2]);
                    m++;
                }
            }
        }
        if (m == 0) {
            return 0;
        }
        ranks1 = computeRanks(tdata1);
        ranks2 = computeRanks(tdata2);
        avgrank = 0.5 * (m - 1); /* Average rank */
        for (int i = 0; i < m; i++) {
            double value1 = ranks1[i];
            double value2 = ranks2[i];
            result += value1 * value2;
            denom1 += value1 * value1;
            denom2 += value2 * value2;
        }
        /* Note: denom1 and denom2 cannot be calculated directly from the number
         * of elements. If two elements have the same rank, the squared sum of
         * their ranks will change.
         */
        result = result * 1.0 / (m * 1.0);
        denom1 = denom1 * 1.0 / (m * 1.0);
        denom2 = denom2 * 1.0 / (m * 1.0);
        result -= avgrank * avgrank;
        denom1 -= avgrank * avgrank;
        denom2 -= avgrank * avgrank;
        if (denom1 <= 0) {
            return 1; /* include '<' to deal with roundoff errors */
        }
        if (denom2 <= 0) {
            return 1; /* include '<' to deal with roundoff errors */
        }
        result = result * 1.0 / Math.sqrt(denom1 * denom2);
        result = 1. - result;
        return result;
    }

    /**
     * Computes Kendall's Tau distance between two rows or two columns.
     * The Kendall distance is defined as one minus Kendall's tau.
     *
     * To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @return double the computed distance
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    public static double kendallsTauCorrelation(double[][] expressionData1,
                                                double[][] expressionData2,
                                                int[][] missingsMask1,
                                                int[][] missingsMask2,
                                                double[] elementsWeights,
                                                int index1,
                                                int index2,
                                                boolean dimension) throws
            Exception {
        checkDimensions(expressionData1, expressionData2, missingsMask1,
                        missingsMask2, elementsWeights, index1, index2,
                        dimension);
        int con = 0;
        int dis = 0;
        int exx = 0;
        int exy = 0;
        boolean flag = false;
        /* flag will remain zero if no nonzero combinations of mask1 and mask2 are
         * found.
         */
        double denomx;
        double denomy;
        double tau;
        int n = (dimension == GENES_DIMENSION) ?
                expressionData1[0].length : expressionData1.length;

        if (dimension == GENES_DIMENSION) {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[index1][i] != 0 &&
                    missingsMask2[index2][i] != 0) {
                    for (int j = 0; j < i; j++) {
                        if (missingsMask1[index1][j] != 0 &&
                            missingsMask2[index2][j] != 0) {
                            double x1 = expressionData1[index1][i];
                            double x2 = expressionData1[index1][j];
                            double y1 = expressionData2[index2][i];
                            double y2 = expressionData2[index2][j];
                            if (x1 < x2 && y1 < y2) {
                                con++;
                            }
                            if (x1 > x2 && y1 > y2) {
                                con++;
                            }
                            if (x1 < x2 && y1 > y2) {
                                dis++;
                            }
                            if (x1 > x2 && y1 < y2) {
                                dis++;
                            }
                            if (x1 == x2 && y1 != y2) {
                                exx++;
                            }
                            if (x1 != x2 && y1 == y2) {
                                exy++;
                            }
                            flag = true;
                        }
                    }
                }
            }
        } else {
            for (int i = 0; i < n; i++) {
                if (missingsMask1[i][index1] != 0 &&
                    missingsMask2[i][index2] != 0) {
                    for (int j = 0; j < i; j++) {
                        if (missingsMask1[j][index1] != 0 &&
                            missingsMask2[j][index2] != 0) {
                            double x1 = expressionData1[i][index1];
                            double x2 = expressionData1[j][index1];
                            double y1 = expressionData2[i][index2];
                            double y2 = expressionData2[j][index2];
                            if (x1 < x2 && y1 < y2) {
                                con++;
                            }
                            if (x1 > x2 && y1 > y2) {
                                con++;
                            }
                            if (x1 < x2 && y1 > y2) {
                                dis++;
                            }
                            if (x1 > x2 && y1 < y2) {
                                dis++;
                            }
                            if (x1 == x2 && y1 != y2) {
                                exx++;
                            }
                            if (x1 != x2 && y1 == y2) {
                                exy++;
                            }
                            flag = true;
                        }
                    }
                }
            }
        }
        if (!flag) {
            return 0.;
        }
        denomx = con + dis + exx;
        denomy = con + dis + exy;
        if (denomx == 0) {
            return 1;
        }
        if (denomy == 0) {
            return 1;
        }
        tau = (con - dis) * 1.0 / Math.sqrt(denomx * denomy);
        return 1. - tau;
    }

    /**
     * Calculates the ranks of the elements in the array
     * <code>dataToRank</code>. Two elements with the same
     * value get the same rank, equal to the average of the
     * ranks had the elements different values.
     *
     * @param dataToRank <code>double[]</code> array that
     *        contains the elements to rank
     * @return <code>double[]</code> the computed ranks for
     *        the given data
     */
    public static double[] computeRanks(double[] dataToRank) {
        int i;
        double[] ranks;
        int[] indexes;
        ranks = new double[dataToRank.length];
        indexes = new int[dataToRank.length];

        /* Call sort to get an index table */
        try {
            sortIndexes(dataToRank, indexes);
        } catch (Exception ex) {
            ex.printStackTrace();
            /* sortIndexes throws an Exception when indexes array doesn't
             have the same size that dataToRank array or if one of these
             arrays are null; none is the case here */
        }
        /* Build a rank table */
        for (i = 0; i < dataToRank.length; i++) {
            ranks[indexes[i]] = i;
        }
        /* Fix for equal ranks */
        i = 0;
        while (i < dataToRank.length) {
            int m;
            double value = dataToRank[indexes[i]];
            int j = i + 1;
            while (j < dataToRank.length && dataToRank[indexes[j]] == value) {
                j++;
            }
            m = j - i; /* number of equal ranks found */
            value = ranks[indexes[i]] + (m - 1) / 2.;
            for (j = i; j < i + m; j++) {
                ranks[indexes[j]] = value;
            }
            i += m;
        }
        return ranks;
    }

    /**
     * Calculates the ranks of the elements in the array
     * <code>dataToRank</code>. Two elements with the same
     * value get the same rank, equal to the average of the
     * ranks had the elements different values.
     *
     * @param dataToRank <code>ArrayList<Double>[]</code>
     *        array that contains the elements to rank
     * @return <code>double[]</code> the computed ranks for
     *        the given data
     * @see #computeRanks(double[] dataToRank)
     */
    public static double[] computeRanks(ArrayList<Double> dataToRank) {
        double[] data = new double[dataToRank.size()];
        Double[] temp = new Double[data.length];
        temp = dataToRank.toArray(temp);
        for (int i = 0; i < data.length; i++) {
            data[i] = temp[i];
        }
        return computeRanks(data);
    }

    /**
     * Sets up an index table given the data, such that data[indexes[i]],
     * with i = 0, ..., data.length is in increasing order. Sorting is
     * done on the indexes. The array data remains unchanged.
     *
     * @param data <code>double[]</code> the data to be sorted
     * @param indexes <code>int[]</code> the indexes of the data to be sorted
     * @throws Exception if <code>data</code> or <code>indexes</code> is
     *         null, or if the length of <code>data</code> and the length
     *         of <code>indexes</code> are not equal
     */
    protected static void sortIndexes(double[] data, int[] indexes) throws
            Exception {
        if (data == null || indexes == null) {
            return;
        }
        if (data.length != indexes.length) {
            throw new Exception("the length of the data and " +
                                "indexes arrays must be equal");
        }
        double[] tempData = new double[data.length];
        for (int i = 0; i < data.length; i++) {
            tempData[i] = data[i];
        }
        Arrays.sort(tempData);
        boolean[] added = new boolean[data.length];
        for (int i = 0; i < added.length; i++) {
            added[i] = false;
        }
        for (int i = 0; i < tempData.length; i++) {
            for (int j = 0; j < data.length; j++) {
                if (data[j] == tempData[i] && !added[j]) {
                    indexes[i] = j;
                    added[j] = true;
                    break;
                }
            }
        }
    }

    /**
     * Checks if the given arguments' comply with each other.
     * <ul>
     * <li><code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions (same number of rows, same number of columns).
     * <li>if dimension == GENES_DIMENSION, then <code>elementsWeights</code>
     * length must equal <code>expressionData1</code> number of rows. Also
     * in this case, index1 must be lower than <code>expressionData1</code>
     * number of rows.
     * <li> if dimension == CONDITIONS_DIMENSION, then
     * <code>elementsWeights</code> length must equal <code>expressionData1</code>
     * number of columns. Also in this case, index2 must be lower than
     * <code>expressionData2</code> number of rows.
     * </ul>
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        operations must be performed; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code>
     *        <li><code>CONDITIONS_DIMENSION</code>
     *        </ul>
     * @throws Exception if the dimensions of the several arrays do not
     *        comply or if the indexes are out of range
     */
    private static void checkDimensions(double[][] expressionData1,
                                        double[][] expressionData2,
                                        int[][] missingsMask1,
                                        int[][] missingsMask2,
                                        double[] elementsWeights,
                                        int index1,
                                        int index2,
                                        boolean dimension) throws Exception {
        if (expressionData1.length != missingsMask1.length ||
            expressionData1[0].length != missingsMask1[0].length ||
            expressionData2.length != missingsMask2.length ||
            expressionData2[0].length != missingsMask2[0].length ||
            expressionData1.length != expressionData2.length ||
            expressionData1[0].length != expressionData2[0].length) {
            throw new Exception("expressionData and missingsMask matrices " +
                                "must have the same sizes");
        }
        /*        if ((dimension == GENES_DIMENSION &&
                     elementsWeights.length != expressionData1.length) ||
                    (dimension == CONDITIONS_DIMENSION &&
                     elementsWeights.length != expressionData1[0].length)) {
         throw new Exception("elementsWeights length must be equal to " +
         "the elements of the rows or columns for " +
                                        "which distance will be measured");
                }
         */
        if ((dimension == GENES_DIMENSION &&
             (index1 >= expressionData1.length ||
              index2 >= expressionData2.length)) ||
            (dimension == CONDITIONS_DIMENSION &&
             (index1 >= expressionData1[0].length ||
              index2 >= expressionData2[0].length))) {
            throw new Exception("indexes are out of range");
        }
    }
}
