package biggests.gui;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

import javax.imageio.*;
import javax.swing.tree.*;

import biggests.tree.*;
import biggests.utils.*;
import de.schlichtherle.io.File;
import de.schlichtherle.io.FileInputStream;
import de.schlichtherle.io.FileOutputStream;

/**
 * <p>Title: BiGGEsTS's Input/Output</p>
 *
 * <p>Description: Provides methods for reading data from and writing
 *                 data to files, for copying, deleting and creating
 *                 directories and files. Restoring and saving
 *                 sessions. Downloading, compressing/decompressing files.
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana Gonalves
 */
public class BiggestsIO {
    /**
     * Some operations may take long. If stop is set to
     * <code>true</code> then such operations should be canceled.
     */
    private static boolean stop = false;

    /**
     * The size of the buffer.
     */
    public static final int BUF_SIZE = 1024;

    /**
     * Image format: PNG
     */
    public static final String PNG = ".png";

    /**
     * Image format: JPG
     */
    public static final String JPG = ".jpg";

    /**
     * Image format: SVG
     */
    public static final String SVG = ".svg";

    /**
     * Maximum image width.
     */
    public static final int MAX_WIDTH = 50000;

    /**
     * Maximum image height.
     */
    public static final int MAX_HEIGHT = 50000;

    /**
     * Sets stop flag.
     *
     * @param st the <code>boolean</code> value
     */
    public static void setStop(boolean st) {
        stop = st;
    }

    /**
     * Returns the value of stop flag.
     *
     * @return the <code>boolean</code> value of stop flag
     */
    public static boolean getStop() {
        return stop;
    }

    /**
     * Creates a set of directories with the specified paths.
     *
     * @param paths the <code>String[]</code> paths of the directories to
     *              be created
     * @throws IOException when some directory could not be created
     */
    public static void initializeFolders(String[] paths) throws IOException {
        for (int i = 0; i < paths.length; i++) {
            File file = new File(paths[i]);
            if (!file.exists()) {
                if (!file.mkdirs()) {
                    throw (new IOException("Could not create " + paths[i] +
                                           " directory"));
                }
            } else {
                if (!file.isDirectory()) {
                    if (!file.delete()) {
                        throw (new IOException("Could not create " + paths[i] +
                                               " directory"));
                    }
                    if (!file.mkdirs()) {
                        throw (new IOException("Could not create " + paths[i] +
                                               " directory"));
                    }
                }
            }
        }
        return;
    }

    /**
     * Deletes a directory content.
     *
     * @param pathToFolder the <code>String</code> path to the specified
     *                     directory
     * @see #deleteRecursively(File[] files)
     */
    public static void deleteFolderContent(String pathToFolder) {
        File folder = new File(pathToFolder);

        if (!folder.exists()) {
            return;
        }
        if (!folder.isDirectory()) {
            folder.delete();
            return;
        }

        File[] folderContent = (File[]) folder.listFiles();
        for (File file : folderContent) {
            file.deleteAll();
        }
    }

    /**
     * Finds for a specific filename in a specific directory path. Returns
     * the absolute path for the file, if it was found, or <code>null</code>
     * othwerwise.
     *
     * @param filename the filename <code>String</code> to search for
     * @param directoryPath the directory path <code>String</code>, where
     *                      to search for the given filename
     * @return the absolute file path <code>String</code>
     * @throws IOException if the directory does not exist
     * @throws Exception if the directory path is not valid
     */
    public static String findFileInDirectoryPath(String filename,
                                                 String directoryPath) throws
            IOException, Exception {
        if (filename == null || directoryPath == null) {
            return null;
        }

        File directory = new File(directoryPath);
        if (!directory.exists()) {
            throw (new IOException("Directory " + directoryPath +
                                   " does not exist."));
        }

        if (!directory.isDirectory()) {
            throw (new Exception("The path " + directoryPath +
                                 " does not correspond to a directory."));
        }

        File file = new File(directoryPath + File.separator + filename);
        if (file.exists() && file.isFile()) {
            return file.getAbsolutePath();
        }

        return null;
    }

    /**
     * Finds the first file with a specific file extension in some
     * specific directory. A file extension starts with ".". Returns the
     * absolute file path for the file, if it was found, or a <code>null</code>
     * pointer othwerwise.
     *
     * @param fileExtension the <code>String</code> which contains the file
     *                      extension to search for; must start with "."
     * @param directoryPath the <code>String</code> which contains the path
     *                      to the directory where to search for the file
     * @return the file path <code>String</code>
     * @throws IOException if the directory does not exist
     * @throws Exception if the directory path or the file extension is not valid
     */
    public static String findFileExtensionInDirectoryPath(String fileExtension,
            String directoryPath) throws IOException, Exception {
        if (fileExtension == null || directoryPath == null) {
            return null;
        }

        File directory = new File(directoryPath);
        if (!directory.exists()) {
            throw (new IOException("Directory " + directoryPath +
                                   " does not exist."));
        }

        if (!directory.isDirectory()) {
            throw (new Exception("The path " + directoryPath +
                                 " does not correspond to a directory."));
        }

        if (!fileExtension.startsWith(".")) {
            throw (new Exception("The suffix " + fileExtension +
                                 " is not a valid file extension."));
        }

        /**
         * search for a file with the specified file extension inside
         * the given directory path
         */
        File[] directoryContent = (File[]) directory.listFiles();
        for (File file : directoryContent) {
            if (file.getName().endsWith(fileExtension) && file.isFile()) {
                return file.getAbsolutePath();
            }
        }
        return null;
    }

    /**
     * Copies a file to another location. If the destination file dores
     * not exist, it is created.
     *
     * @param source the <code>File</code> to copy to another location
     * @param destination the <code>File</code> which is the destination target
     * @throws IOException when it is not possible to copy the file
     */
    public static void copy(File source, File destination) throws IOException {
        InputStream in = new FileInputStream(source);
        OutputStream out = new FileOutputStream(destination);

        /**
         * copy bytes from the source file to the destination file
         */
        byte[] buffer = new byte[BUF_SIZE];
        int length;
        while ((length = in.read(buffer)) > 0) {
            out.write(buffer, 0, length);
        }
        in.close();
        out.close();
    }

    /**
     * Reads gene expression data from a txt file.
     * Some rows may contain conditions' names and some columns may
     * contain genes' IDs.
     *
     * @param file <code>String</code> the filepath which contains
     *        the file to read
     * @param separator the <code>char</code>acter that delimits the data
     * @param missingValue <code>float</code> the value to fill missings
     * @return Object[][] the expression data read from the given file
     * @throws FileNotFoundException
     * @throws IOException
     */
    public static Object[][] readExpressionDataFromFileTXT(
            String file,
            char separator,
            float missingValue) throws FileNotFoundException, IOException {

        // EXCEPTION!!!
        // 1) separator can only be '\t', ';' ',' or ' '
        // 2) the file type is .txt

        if (separator != '\t' && separator != ' ' && separator != ';') {
            throw new IOException(
                    "Invalid separator: separator can only be '\t', ';' ',' or ' '.");
        }

        if (!file.toLowerCase().endsWith(".txt")) {
            throw new IOException(
                    "Invalid file type: file type can only be '.txt'.");
        }

        // FileNotFoundException
        BufferedReader br = new BufferedReader(new FileReader(file));

        String line = br.readLine();

        StringTokenizer st = new StringTokenizer(line,
                                                 new Character(separator).
                                                 toString());
        int numberOfTokens = st.countTokens();
        if (numberOfTokens == 0 || numberOfTokens == 1) { //separator != file separator
            throw new IOException("Invalid separator: separator found in " +
                                  "the file is different from the given one.");
        }

        int numberOfRows = 0;

        br.close();

        // FileNotFoundException
        br = new BufferedReader(new FileReader(file));
        while ((line = br.readLine()) != null) {
            numberOfRows++;
        }

        Object[][] expressionMatrix = new Object[numberOfRows][numberOfTokens];

        br.close();
        br = new BufferedReader(new FileReader(file));
        for (int i = 0; i < expressionMatrix.length; i++) {
            line = br.readLine();
            for (int x = 0; x < expressionMatrix[0].length; x++) {
                line = line.replaceAll("\t\t", "\t@\t");
            }
            st = new StringTokenizer(line,
                                     new Character(separator).toString());

            for (int j = 0; j < expressionMatrix[0].length; j++) {
                String s = null;
                try {
                    s = st.nextToken();
                } catch (Exception ex) {
                    expressionMatrix[i][j] = Float.valueOf(missingValue);
                }
                if (s == null || s.equals("") || s.equals("@")) {
                    expressionMatrix[i][j] = Float.valueOf(missingValue);
                } else {
                    try {
                        expressionMatrix[i][j] = Float.parseFloat(s);
                    } catch (NumberFormatException ex) {
                        expressionMatrix[i][j] = s;
                    }
                }
            }
        }

        return expressionMatrix;
    }

    /**
     * Reads conversion file entries from a text file. Conversion entries
     * must be in a two column format, for which the first element is the
     * probeset ID and the second one is the GO gene ID, both delimited
     * by the <code>separator</code> character.
     *
     * @param filePath <code>String</code> the path of the file with
     *        conversion entries to read
     * @param separator the <code>char</code>acter that delimits the data
     * @return String[][] the conversion entries read from the given file
     * @throws FileNotFoundException
     * @throws IOException
     */
    public static String[][] readConversionEntriesFromFileTXT(String filePath,
            char separator) throws FileNotFoundException, IOException {
        // EXCEPTION!!!
        // 1) separator can only be '\t', ';' ',' or ' '
        // 2) the file type is .txt

        if (separator != '\t' && separator != ' ' && separator != ';') {
            throw new IOException(
                    "Invalid separator: separator can only be '\t', ';' ',' or ' '.");
        }

        if (!filePath.toLowerCase().endsWith(".txt")) {
            throw new IOException(
                    "Invalid file type: file type can only be '.txt'.");
        }

        // FileNotFoundException
        BufferedReader br = new BufferedReader(new FileReader(filePath));

        String line = br.readLine();

        StringTokenizer st = new StringTokenizer(line,
                                                 new Character(separator).
                                                 toString());
        int numberOfTokens = st.countTokens();
        if (numberOfTokens == 0) { //separator != file separator
            throw new IOException("Invalid separator: separator found in " +
                                  "the file is different from the given one.");
        }
        if (numberOfTokens != 2) {
            throw new IOException("Invalid data format.");
        }

        int numberOfRows = 0;

        br.close();

        // FileNotFoundException
        br = new BufferedReader(new FileReader(filePath));
        while ((line = br.readLine()) != null) {
            numberOfRows++;
        }

        br.close();

        br = new BufferedReader(new FileReader(filePath));
        String[][] conversionEntries = new String[numberOfRows][2];
        int i = 0;
        while ((line = br.readLine()) != null) {
            st = new StringTokenizer(line, new Character(separator).toString());
            if (numberOfTokens != 2) {
                throw new IOException("Invalid data format.");
            }
            conversionEntries[i][0] = st.nextToken();
            conversionEntries[i][1] = st.nextToken();
            i++;
        }
        return conversionEntries;
    }

    /**
     * Saves a session to a .zip archive file.
     *
     * @param tree the <code>DatasetTree</code> which contains datasets
     *             loaded and manipulated in the session
     * @param pathToFunctionResults <code>String</code> the path to directory
     *             which contains the files resulting from function analysis
     * @param pathToSessionArchive <code>String</code> the path to the
     *             resulting archive file
     * @throws NotSerializableException if one of the objects that system
     *         tried to write to a file is not serializable
     * @throws FileNotFoundException if files could not be found
     * @throws IOException if could not write/read from files
     */
    public static void saveSession(DatasetTree tree,
                                   String pathToFunctionResults,
                                   String pathToSessionArchive) throws
            NotSerializableException, FileNotFoundException, IOException {
        File treeFile = new File("tree.big");

        ObjectOutputStream oos = new ObjectOutputStream(new
                FileOutputStream(treeFile));

        oos.writeObject(tree);
        oos.flush();
        oos.writeObject(tree.getSelectionPath());
        oos.flush();
        oos.close();

        treeFile.archiveCopyTo(new File(pathToSessionArchive + File.separator +
                                        treeFile.getName()));
        File sessionResults = new File(pathToFunctionResults);
        sessionResults.archiveCopyAllTo(new File(pathToSessionArchive +
                                                 File.separator +
                                                 sessionResults.getName()));

        treeFile.delete();
        return;
    }

    /**
     * Saves a session to a .zip archive file.
     *
     * @param tree the <code>DatasetTree</code> which contains datasets
     *             loaded and manipulated in the session
     * @param pathToFunctionResults <code>String</code> the path to directory
     *             which contains the files resulting from function analysis
     * @param pathToSessionArchive <code>String</code> the path to the
     *             resulting archive file
     * @throws NotSerializableException if one of the objects that system
     *         tried to write to a file is not serializable
     * @throws FileNotFoundException if files could not be found
     * @throws IOException if could not write/read from files
     */
    public static void saveSession2(DatasetTree tree,
                                   String pathToFunctionResults,
                                   String pathToSessionArchive) throws
            NotSerializableException, FileNotFoundException, IOException {
        ArrayList<Integer> counts = new ArrayList<Integer>();
        File treeFile = new File("tree.big");
        File countsFile = new File("counts.txt");

        ObjectOutputStream outputTree = new ObjectOutputStream(new
                FileOutputStream(treeFile));
        PrintWriter outputCounts = new PrintWriter(countsFile);

        Enumeration<DefaultMutableTreeNode> datasets = tree.getRoot().children();
        if (datasets == null) {
            outputCounts.print(0);
        }
        else {
            int numberOfDatasets = 0;
            while (datasets.hasMoreElements()) {
                numberOfDatasets++;
                DefaultMutableTreeNode dataset = datasets.nextElement();
                Enumeration<DefaultMutableTreeNode> datasetChildren =
                        dataset.children();
                // write the original expression matrix
                outputTree.writeObject(
                        datasetChildren.nextElement().getUserObject());

                // check the preprocessed matrices
                Enumeration<DefaultMutableTreeNode> preprocessed =
                        datasetChildren.nextElement().children();
                int preprocessedCount = 0;
                if (preprocessed != null) {
                    while (preprocessed.hasMoreElements()) {
                        preprocessedCount++;
                        outputTree.writeObject(preprocessed
                                               .nextElement().getUserObject());
                    }
                }

                // check the discretized matrices
                Enumeration<DefaultMutableTreeNode> discretized =
                        datasetChildren.nextElement().children();
                int discretizedCount = 0;
                if (discretized != null) {
                    while (discretized.hasMoreElements()) {
                        discretizedCount++;
                        outputTree.writeObject(discretized
                                               .nextElement().getUserObject());
                    }
                }

                // check the groups of biclusters
                Enumeration<DefaultMutableTreeNode> groupsOfBiclusters =
                        datasetChildren.nextElement().children();
                int groupsOfBiclustersCount = 0;
                if (groupsOfBiclusters != null) {
                    while (groupsOfBiclusters.hasMoreElements()) {
                        groupsOfBiclustersCount++;
                        outputTree.writeObject(groupsOfBiclusters
                                               .nextElement().getUserObject());
                    }
                }

                // check the post-processed groups of biclusters
                Enumeration<DefaultMutableTreeNode> postprocessed =
                        datasetChildren.nextElement().children();
                int postprocessedCount = 0;
                if (postprocessed != null) {
                    while (postprocessed.hasMoreElements()) {
                        postprocessedCount++;
                        outputTree.writeObject(postprocessed
                                   .nextElement().getUserObject());
                    }
                }

                counts.add(preprocessedCount);
                counts.add(discretizedCount);
                counts.add(groupsOfBiclustersCount);
                counts.add(postprocessedCount);
            }

            outputCounts.println(numberOfDatasets);

            for (int i = 0; i < numberOfDatasets*4; i++) {
                outputCounts.println(counts.get(i));
            }
        }

        outputCounts.close();
        outputTree.close();

        treeFile.archiveCopyTo(new File(pathToSessionArchive + File.separator +
                                        treeFile.getName()));
        countsFile.archiveCopyTo(new File(pathToSessionArchive + File.separator +
                                        countsFile.getName()));
        File sessionResults = new File(pathToFunctionResults);
        sessionResults.archiveCopyAllTo(new File(pathToSessionArchive +
                                                 File.separator +
                                                 sessionResults.getName()));


        treeFile.delete();
        countsFile.delete();
        return;
    }


    /**
     * Restores a session from a .zip archive file. Zip file must contain
     * a file named tree.big with an object of type <code>DatasetTree</code>
     * and a directory named <code>Function_Results</code> which contains
     * resulting files from function analysis.
     *
     * @param frame the BiGGEsTS main window
     * @param pathToSessionArchive <code>String</code> the path to the zip
     *              archive which contains the files of the session to restore
     * @throws Exception if the format of the session files could not be recognized
     */
    public static void restoreSession(BiggestsFrame frame,
                                      String pathToSessionArchive) throws
            Exception {
        File sessionArchive = new File(pathToSessionArchive);
        File sessionTree = new File(pathToSessionArchive + File.separator +
                                    "tree.big");
        File sessionResults = new File(pathToSessionArchive + File.separator +
                                       "Function_Results");

        if (!sessionArchive.isArchive()) {
            throw new Exception("Session file is not a zip archive.");
        }
        if (!sessionTree.exists() || !sessionTree.isFile() ||
            !sessionResults.exists() || !sessionResults.isDirectory()) {
            throw new Exception("Unrecognized session format.");
        }

        ObjectInputStream ois = new ObjectInputStream(new FileInputStream(
                sessionTree));
        frame.setTree((DatasetTree) ois.readObject());

        File functionResults = new File(frame.getPathToFunctionResults());
        functionResults.deleteAll();

        functionResults.copyAllFrom(sessionResults);
        frame.getTree().setSelectionPath((TreePath) ois.readObject());
        ois.close();
    }

    /**
     * Restores a session from a .zip archive file. Zip file must contain
     * a file named tree.big with an object of type <code>DatasetTree</code>
     * and a directory named <code>Function_Results</code> which contains
     * resulting files from function analysis.
     *
     * @param frame the BiGGEsTS main window
     * @param pathToSessionArchive <code>String</code> the path to the zip
     *              archive which contains the files of the session to restore
     * @throws Exception if the format of the session files could not be recognized
     */
    public static void restoreSession2(BiggestsFrame frame,
                                      String pathToSessionArchive) throws
            Exception {
        File sessionArchive = new File(pathToSessionArchive);
        File sessionTree = new File(pathToSessionArchive + File.separator +
                                    "tree.big");
        File countsFile = new File(pathToSessionArchive + File.separator +
                                    "counts.txt");
        File sessionResults = new File(pathToSessionArchive + File.separator +
                                       "Function_Results");

        if (!sessionArchive.isArchive()) {
            throw new Exception("Session file is not a zip archive.");
        }
        if (!sessionTree.exists() || !sessionTree.isFile() ||
            !sessionResults.exists() || !sessionResults.isDirectory()) {
            throw new Exception("Unrecognized session format.");
        }

        BufferedReader inputCounts =
            new BufferedReader(new de.schlichtherle.io.FileReader(countsFile));
        ObjectInputStream inputTree = new ObjectInputStream(
            new FileInputStream(sessionTree));
        int numberOfDatasets = Integer.parseInt(inputCounts.readLine());
        if (numberOfDatasets == 0) {
            inputCounts.close();
            inputTree.close();
            return;
        }

        frame.setTree(new DatasetTree());
        DatasetTree tree = frame.getTree();

        for (int i = 0; i < numberOfDatasets; i++) {
            // read and add the original matrix
            tree.addLoadedDataset((NodeInfo)inputTree.readObject(), true);
            TreePath pathToDataset = tree.getSelectionPath();

            // check other matrices and biclusters
            for (int k = 0; k < 4; k++) {
                int tempCount = Integer.parseInt(inputCounts.readLine());
                for (int j = 0; j < tempCount; j++) {
                    tree.addNewNode((NodeInfo)inputTree.readObject(),
                                    pathToDataset, false);
                }
            }
        }

        DefaultMutableTreeNode root = tree.getRoot();
        Enumeration<Object> children = root.children();
        while (children.hasMoreElements()) {
            tree.expandPath(new TreePath(
                    new Object[] {root, children.nextElement()}));
        }

        tree.setSelectionPath(new TreePath(
            new Object[]{root, root.getFirstChild(),
            ((DefaultMutableTreeNode)root.getFirstChild()).getFirstChild()}));

        File functionResults = new File(frame.getPathToFunctionResults());
        functionResults.deleteAll();

        functionResults.copyAllFrom(sessionResults);

        inputCounts.close();
        inputTree.close();
    }

    /**
     * Downloads gene ontology file.
     *
     * @param geneOntologyFilesURL <code>String</code> the URL where
     *        gene ontology files can be found
     * @param pathToGOFiles <code>String</code> the path to gene ontology
     *        files directory/folder
     * @param geneOntologyFileName <code>String</code> the name of the
     *        gene ontology file to be downloaded (usually gene_ontology.obo)
     * @param progressFrame <code>ProgressBarFrame</code> the progress bar to
     *        send information about the progress of the operation
     * @throws Exception
     */
    public static void downloadGeneOntologyFile(String geneOntologyFilesURL,
                                                String pathToGOFiles,
                                                String geneOntologyFileName,
                                                ProgressBarFrame progressFrame) throws
            Exception {
        File temp;
        String addressToGeneOntologyFile = geneOntologyFilesURL
                                           + geneOntologyFileName;
        String pathToGeneOntologyFile = pathToGOFiles +
                                        File.separator + geneOntologyFileName;
        File geneOntologyFile = new File(pathToGeneOntologyFile);

        temp = new File(pathToGOFiles + File.separator + "go.obo");
        geneOntologyFile.renameTo(temp);

        String message = "downloading gene ontology file...";
        try {
            download(addressToGeneOntologyFile,
                     pathToGeneOntologyFile, progressFrame, message);
            temp.delete();
        } catch (Exception ex) {
            geneOntologyFile.delete();
            temp.renameTo(geneOntologyFile);
            ex.printStackTrace();
            throw new Exception(ex.getMessage());
        }
    }

    /**
     * Downloads gene association file.
     *
     * @param geneAssociationFilesURL <code>String</code> the URL where
     *        gene association files can be found
     * @param pathToGOFiles <code>String</code> the path to gene ontology
     *        files directory/folder
     * @param geneAssociationFileName <code>String</code> the name of the
     *        gene ontology file to be downloaded (usually gene_ontology.obo)
     * @param progressFrame <code>ProgressBarFrame</code> the progress bar to
     *        send information about the progress of the operation
     * @throws Exception
     */
    public static void downloadGeneAssociationFile(
            String geneAssociationFilesURL,
            String pathToGOFiles,
            String geneAssociationFileName,
            ProgressBarFrame progressFrame) throws Exception {
        String associationFileRevision = "?rev=HEAD";
        File temp;

        String addressToGeneAssociationFileGZ = geneAssociationFilesURL
                                                + geneAssociationFileName +
                                                ".gz"
                                                + associationFileRevision;
        String pathToGeneAssociationFile = pathToGOFiles +
                                           File.separator +
                                           geneAssociationFileName;
        String pathToGeneAssociationFileGZ = pathToGeneAssociationFile + ".gz";

        File geneAssociationFile = new File(pathToGeneAssociationFile);

        temp = new File(pathToGOFiles + File.separator + "ga.org");
        File geneAssociationFileGZ = new File(pathToGeneAssociationFile + ".gz");
        geneAssociationFile.renameTo(temp);

        String message;
        try {
            message = "downloading gene association file...";
            progressFrame.setString(message + " 0 kB");
            download(addressToGeneAssociationFileGZ,
                     pathToGeneAssociationFileGZ, progressFrame, message);
        } catch (Exception ex) {
            geneAssociationFileGZ.delete();
            temp.renameTo(geneAssociationFile);
            ex.printStackTrace();
            throw new Exception(ex.getMessage());
        }

        progressFrame.setString("unzipping gene association file...");
        try {
            gunzipFile(geneAssociationFileGZ, new File(pathToGOFiles));
            geneAssociationFileGZ.delete();
            temp.delete();
        } catch (Exception ex) {
            geneAssociationFileGZ.delete();
            geneAssociationFile.delete();
            temp.renameTo(geneAssociationFile);
            ex.printStackTrace();
            throw new Exception(ex.getMessage());
        }
    }

    /**
     * Downloads the file from the Internet located in the specified
     * <code>address</code>. The downloaded file is stored in
     * <code>localFilePath</code> (this path already contains the name
     * of the file).
     *
     * @param address <code>String</code> the address of the file to download
     * @param localFilePath <code>String</code> the local file path to store
     *        the downloaded file
     * @param progressFrame ProgressBarFrame
     * @param message String
     * @throws IOException if input and/or output stream could not be closed
     * @throws Exception if file could not be downloaded
     */
    public static void download(String address,
                                String localFilePath,
                                ProgressBarFrame progressFrame,
                                String message) throws IOException, Exception {
        OutputStream out = null;
        URLConnection conn = null;
        InputStream in = null;

        try {
            if (stop) {
                throw new Exception("Operation canceled.");
            }
            URL url = new URL(address);
            out = new BufferedOutputStream(new FileOutputStream(new File(
                    localFilePath)));
            conn = url.openConnection();
            in = conn.getInputStream();
            byte[] buffer = new byte[BUF_SIZE];
            int numRead;
            double bytesRead = 0;
            NumberFormat nf = NumberFormat.getInstance(Locale.ENGLISH);
            nf.setMaximumFractionDigits(0);
            nf.setGroupingUsed(false);
            /*            while (((numRead = in.read(buffer)) != -1) && (bytesRead<1000) && (!stop)) {
                            bytesRead += numRead*1.0/1024;
             progressFrame.setString(message + " " + nf.format(bytesRead) + " kB");
                            out.write(buffer, 0, numRead);
                            out.flush();
                        }
                        if (numRead == -1) {
                            out.flush();
                            out.close();
                            in.close();
                            return;
                        }
                        bytesRead = bytesRead*1.0/1024;
             */
            nf.setMaximumFractionDigits(3);
            nf.setMinimumFractionDigits(3);
            while (((numRead = in.read(buffer)) != -1) && (!stop)) {
                bytesRead += numRead * 1.0 / 1024 / 1024;
                progressFrame.setString(message + " " + nf.format(bytesRead) +
                                        " MB");
                out.write(buffer, 0, numRead);
                out.flush();
            }
            if (stop) {
                stop = false;
                throw new Exception("Operation canceled.");
            }
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
                if (out != null) {
                    out.close();
                }
            } catch (IOException ioe) {
                ioe.printStackTrace();
                throw new IOException(ioe.getMessage());
            }
        }
    }

    /**
     * Gunzips the input archive. Sends the output to the directory specified
     * by dir_output. Assumes that the input file name ends with ".gz"
     *
     * @param file_input the input gzipped <code>File</code>
     * @param dir_output <code>File</code> directory to send the gunzipped file to
     * @throws IOException
     * @throws Exception
     */
    public static void gunzipFile(File file_input, File dir_output) throws
            IOException, Exception {

        // Create a buffered gzip input stream to the archive file.
        GZIPInputStream gzip_in_stream;
        FileInputStream in = new FileInputStream(file_input);
        BufferedInputStream source = new BufferedInputStream(in);
        gzip_in_stream = new GZIPInputStream(source);

        // Use the name of the archive for the output file name but
        // with ".gz" stripped off.
        String file_input_name = file_input.getName();
        String file_output_name =
                file_input_name.substring(0, file_input_name.length() - 3);

        // Create the decompressed output file.
        File output_file = new File(dir_output, file_output_name);

        // Decompress the gzipped file by reading it via
        // the GZIP input stream. Will need a buffer.
        byte[] input_buffer = new byte[BUF_SIZE];
        int len = 0;
        // Create a buffered output stream to the file.
        FileOutputStream out = new FileOutputStream(output_file);
        BufferedOutputStream destination =
                new BufferedOutputStream(out, BUF_SIZE);

        // Now read from the gzip stream, which will decompress the data,
        // and write to the output stream.
        while (((len = gzip_in_stream.read(input_buffer, 0, BUF_SIZE)) != -1) &&
                       (!stop)) {
            destination.write(input_buffer, 0, len);
        }
        destination.flush(); // Ensure that all data is written to the output.
        out.close();
        destination.close();

        gzip_in_stream.close();

        if (stop) {
            throw new Exception("Operation canceled");
        }
    }

    /**
     * Exports a GUI component to an image file. Supported image
     * file types are: PNG and JPG. The image will be created with
     * the dimensions of the original component's image.
     * The path of the image file is given in <code>filePath</code>
     * and must contain the appropriate file extension.
     *
     * @param component <code>Component</code> the component to export to image
     * @param format <code>String</code> the format of the image (PNG or JPG)
     * @param filePath <code>String</code> the path to the image file
     * @throws IOException if the component cannot be exported to the image file
     */
    public static void exportComponentToImage(Component component,
                                              String format,
                                              String filePath) throws
            IOException {
        if (!format.equals(PNG) && !format.equals(JPG)) {
            throw new IOException("Invalid file format.");
        }
        BufferedImage image = null;
        int width = component.getWidth();
        int height = component.getHeight();
        image = new BufferedImage(
                width, height,
                BufferedImage.TYPE_INT_RGB);
        if (width > MAX_WIDTH) {
            width = MAX_WIDTH;
        }
        if (height > MAX_HEIGHT) {
            height = MAX_HEIGHT;
        }
        image = image.getSubimage(0, 0, width, height);
        component.paint(image.getGraphics());
        if (format.equals(PNG)) {
            ImageIO.write(image, "png", new File(filePath));
            return;
        }
        else if (format.equals(JPG)) {
            ImageIO.write(image, "jpg", new File(filePath));
        }
    }

    /**
     * Exports a GUI component to an image file. Supported image
     * file types are: PNG and JPG. The image will be created from
     * <code>x</code>- and <code>y</code>-coordinates (left-most
     * top corner), with <code>width</code> width and
     * <code>height</code> height from the original component's image.
     * The path of the file is given in <code>filePath</code> and must
     * contain the appropriate file extension.
     *
     * @param component <code>Component</code> the component to export to image
     * @param x <code>int</code> the initial <code>x</code>-coordinate
     *          of the original image from which the final image will
     *          be produced (left-most top corner)
     * @param y <code>int</code> the initial <code>y</code>-coordinate
     *          of the original image from which the final image will
     *          be produced (left-most top corner)
     * @param width <code>int</code> the width of the image to produce
     * @param height <code>int</code> the height of the image to produce
     * @param format <code>String</code> the format of the image (PNG or JPG)
     * @param filePath <code>String</code> the path to the image file
     * @throws IOException if the component cannot be exported to the image file
     */
    public static void exportComponentToImage(Component component,
                                              int x,
                                              int y,
                                              int width,
                                              int height,
                                              String format,
                                              String filePath) throws
            IOException {
        if (!format.equals(PNG) && !format.equals(JPG)) {
            throw new IOException("Invalid file format.");
        }
        BufferedImage image = null;
        image = new BufferedImage(
                width, height,
                BufferedImage.TYPE_INT_RGB);
        if (width > MAX_WIDTH) {
            width = MAX_WIDTH;
        }
        if (height > MAX_HEIGHT) {
            height = MAX_HEIGHT;
        }
        if (x+width > image.getRaster().getWidth()) {
            x = 0;
        }
        if (y+height > image.getRaster().getHeight()) {
            y = 0;
        }
        image = image.getSubimage(x, y, width, height);
        component.paint(image.getGraphics());
        if (format.equals(PNG)) {
            ImageIO.write(image, "png", new File(filePath));
            return;
        }
        else if (format.equals(JPG)) {
            ImageIO.write(image, "jpg", new File(filePath));
        }
    }
}
