package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import biggests.tree.*;
import de.schlichtherle.io.*;
import smadeira.biclustering.*;

/**
 * <p>Title: Post-Processing Panel</p>
 *
 * <p>Description: Displays options to select and apply post-processing
 *                 techniques to groups of biclusters.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana Gonalves
 */
public class PanelPostProcessing extends PanelBasicIOValidation {
    /*************************************************************************
     * ***********************************************************************
     * ***********************************************************************
     *                      FILTERING CRITERIA INDEXES
     * ***********************************************************************
     * ***********************************************************************
     * ***********************************************************************
     */
    /**
     * The index of filter trivial biclusters option.
     */
    public static final int FILTER_TRIVIAL = 0;

    /**
     * The index of 'filter biclusters with less than a given
     * number of genes' option.
     */
    public static final int FILTER_GENES = 1;

    /**
     * The index of 'filter biclusters with less than a given number of
     * conditions' option.
     */
    public static final int FILTER_CONDITIONS = 2;

    /**
     * The index of 'filter biclusters with less than a given
     * number of genes and less than a given number of conditions'
     * option.
     */
    public static final int FILTER_GENES_CONDITIONS = 3;

    /**
     * The index of 'filter biclusters with an ACV value below a
     * given threshold' option.
     */
    public static final int FILTER_ACV = 4;

    /**
     * The index of 'filter biclusters with an ARV value below a given
     * threshold' option.
     */
    public static final int FILTER_ARV = 5;

    /**
     * The index of 'filter trivial biclusters and biclusters
     * with ACV and ARV values below given thresholds' option.
     */
    public static final int FILTER_ACV_ARV = 6;

    /**
     * The index of 'filter biclusters with an MSR value below
     * a given threshold' option.
     */
    public static final int FILTER_MSR = 7;

    /**
     * The index of 'filter biclusters smaller than a given size' option.
     */
    public static final int FILTER_SIZE = 8;

    /**
     * The index of 'filter biclusters smaller than a given size and
     * with ARV and ACV values below given thresholds' option.
     */
    public static final int FILTER_SIZE_ARV_ACV = 9;

    /**
     * The index of 'filter biclusters with less than given numbers
     * of genes and conditions and with ARV and ACV values below
     * given thresholds' option.
     */
    public static final int FILTER_CONDITIONS_GENES_ARV_ACV = 10;

    /**
     * The index of 'filter biclusters with less than given numbers
     * of genes and conditions and with ARV, ACV and MSR values below
     * given thresholds' option.
     */
    public static final int FILTER_CONDITIONS_GENES_ARV_ACV_MSR = 11;

    /**
     * The index of 'filter biclusters with a row similarity below
     * a given threshold' option.
     */
    public static final int FILTER_OVERLAPPING_GENES = 12;

    /**
     * The index of 'filter biclusters with row and column similarities
     * below given thresholds' option.
     */
    public static final int FILTER_OVERLAPPING_GENES_CONDITIONS = 13;

    /**
     * The index of 'filter biclusters with constant pattern' option.
     */
    public static final int FILTER_CONSTANT_PATTERN = 14;

    public static final int FILTER_PATTERN_PVALUE_NOT_SIGNIFICANT = 15;

    public static final int FILTER_PATTERN_COLUMNS_PVALUE_NOT_SIGNIFICANT = 16;

    /**************************************************************************
     * ************************************************************************
     * ************************************************************************
     *                     SORTING CRITERIA INDEXES
     * ************************************************************************
     * ************************************************************************
     * ************************************************************************
     */
    /**
     * The index of 'sort biclusters by their ACV values' option.
     */
    public static final int SORT_ACV = 0;

    /**
     * The index of 'sort biclusters by their ARV values' option.
     */
    public static final int SORT_ARV = 1;

    /**
     * The index of 'sort biclusters by their MSR values' option.
     */
    public static final int SORT_MSR = 2;

    /**
     * The index of 'sort biclusters by the differences between
     * their ARV and ACV values' option.
     */
    public static final int SORT_ARV_ACV_DIFF = 3;

    /**
     * The index of 'sort biclusters by the differences between
     * their ARV and MSR values' option.
     */
    public static final int SORT_ARV_MSR_DIFF = 4;

    /**
     * The index of 'sort biclusters by their number of genes' option.
     */
    public static final int SORT_GENES = 5;

    /**
     * The index of 'sort biclusters by their number of conditions' option.
     */
    public static final int SORT_CONDITIONS = 6;

    /**
     * The index of 'sort biclusters by their size' option.
     */
    public static final int SORT_SIZE = 7;

    /**
     * The index of 'sort biclusters by their variance' option.
     */
    public static final int SORT_VARIANCE = 8;

    /**
     * The index of 'sort biclusters by their best GO p-value' option.
     */
    public static final int SORT_BEST_GO_PVALUE = 9;

    /**
     * The index of 'sort biclusters by their best GO Bonferroni
     * corrected p-value' option.
     */
    public static final int SORT_BEST_GO_BONFERRONI_PVALUE = 10;

    /**
     * The index of 'sort biclusters by their pattern p-value' option.
     */
    public static final int SORT_PATTERN_PVALUE = 11;

    /**
     * The index of 'sort biclusters by the p-value for their pattern
     * with columns' option.
     */
    public static final int SORT_PATTERN_COLUMNS_PVALUE = 12;


    /**
     * Number of genes parameter name.
     */
    public static final String PARAMETER_GENES = "number of genes";

    /**
     * Number of conditions parameter name.
     */
    public static final String PARAMETER_CONDITIONS = "number of conditions";

    /**
     * Size parameter name.
     */
    public static final String PARAMETER_SIZE = "size";

    /**
     * Similarity parameter name.
     */
    public static final String PARAMETER_SIMILARITY = "similarity";

    /**
     * Average column variance parameter name.
     */
    public static final String PARAMETER_ACV = "average column variance";

    /**
     * Average row variance parameter name.
     */
    public static final String PARAMETER_ARV = "average row variance";

    /**
     * Mean squared residue parameter name.
     */
    public static final String PARAMETER_MSR = "mean squared residue";

    /**
     * Markov chain order parameter name.
     */
    public static final String PARAMETER_MARKOV_FILTER = "Markov chain order to filter";

    public static final String PARAMETER_MARKOV_SORT = "Markov chain order to sort";

    public static final String PARAMETER_PVALUE = "p-value threshold";

    /**
     * Bicluster group object to post-process.
     */
    private Biclustering biclustering;

    /**
     * The path of bicluster group object to post-process in the dataset tree.
     */
    private TreePath selectionPath;

    /**
     * The layout of this panel.
     */
    private SpringLayout springLayout;

    /**
     * Filter check box. Allows the user to select/deselect filter
     * post-processing option.
     */
    private JCheckBox filterCheckBox;

    /**
     * Filter criteria label. Displays a title for filter criteria list.
     */
    private JLabel filterCriteriaLabel;

    /**
     * Filter criteria list. Contains available criteria to filter
     * biclusters of a biclusters group.
     */
    private JList filterCriteriaList;

    /**
     * Sort check box. Allows the user to select/deselect sort
     * post-processing option.
     */
    private JCheckBox sortCheckBox;

    /**
     * Sort criteria label. Displays a title for sort criteria list.
     */
    private JLabel sortCriteriaLabel;

    /**
     * Sort criteria list. Contains available criteria to sort biclusters
     * of a biclusters group.
     */
    private JList sortCriteriaList;

    /**
     * Filter description label. Displays a description of the selected
     * filter criteria and the text fields so the user can enter
     * filter parameters.
     */
    private JLabel filterDescriptionLabel;

    /**
     * Sort description label. Displays a description of the selected
     * sort criteria and the text fields so the user can enter
     * sort parameters.
     */
    private JLabel sortDescriptionLabel;

    /**
     * ACV text field. Allows the user to input average column variance threshold.
     */
    private JTextField acvTextField;

    /**
     * ARV text field. Allows the user to input average row variance threshold.
     */
    private JTextField arvTextField;

    /**
     * MSR text field. Allows the user to input mean squared residue threshold.
     */
    private JTextField msrTextField;

    /**
     * P-value text field. Allows the user to input the p-value threshold.
     */
    private JTextField pValueTextField;

    /**
     * Number of conditions text field. Allows the user to input
     * gene number threshold.
     */
    private JTextField conditionsTextField;

    /**
     * Number of genes text field. Allows the user to input
     * gene number threshold.
     */
    private JTextField genesTextField;

    /**
     * Similarity text field. Allows the user to input
     * similarity threshold.
     */
    private JTextField similarityTextField;

    /**
     * Size text filed. Allows the user to input size threshold.
     */
    private JTextField sizeTextField;

    private JTextField markovChainOrderFilterTextField;

    /**
     * Markov chain order text field. Allows the user to input Markov chain order.
     */
    private JTextField markovChainOrderSortTextField;

    /**
     * Filter panel. Contains information about filtering options.
     */
    private JPanel filterPanel;

    /**
     * The layout of the filter panel.
     */
    private SpringLayout filterPanelLayout;

    /**
     * Sort panel. Contains information about soring options.
     */
    private JPanel sortPanel;

    /**
     * The layout of the sort panel.
     */
    private SpringLayout sortPanelLayout;

    /**
     * Ascending sorting button. Allows the user to select sorting
     * by the specified parameters in ascending order.
     */
    private JRadioButton ascendingButton;

    /**
     * Descending sorting button. Allows the user to select sorting
     * by the specified parameters in descending order.
     */
    private JRadioButton descendingButton;

    /**
     * Button group. Groups ascending and descending sorting buttons, to
     * ensure that only one of them is selected at any time.
     */
    private ButtonGroup group;

    /**
     * Apply button. When clicked, applies post-processing options
     * to the selected biclusters group.
     */
    private JButton applyButton;

    /**
     * The application's main frame.
     */
    private BiggestsFrame mainFrame;

    /**
     * Creates a new post-processing panel.
     *
     * @param frame the main <code>BiggestsFrame</code>
     */
    public PanelPostProcessing(BiggestsFrame frame) {
        mainFrame = frame;
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Component initialization.
     *
     * @throws Exception if components could not be initialized
     */
    private void jbInit() throws Exception {
        this.setPreferredSize(new Dimension(760, 630));
        springLayout = new SpringLayout();
        this.setLayout(springLayout);
        this.setBorder(BorderFactory.createEmptyBorder(50, 50, 50, 50));

        acvTextField = new JTextField(3);
        arvTextField = new JTextField(3);
        msrTextField = new JTextField(3);
        conditionsTextField = new JTextField(3);
        genesTextField = new JTextField(3);
        similarityTextField = new JTextField(3);
        sizeTextField = new JTextField(3);
        markovChainOrderFilterTextField = new JTextField(3);
        markovChainOrderSortTextField = new JTextField(3);
        pValueTextField = new JTextField(3);

        filterCheckBox = new JCheckBox("Filter");
        filterCheckBox.addItemListener(new
                                       PanelPostProcessing_filterCheckBox_itemAdapter(this));
        filterCriteriaLabel = new JLabel("Filtering Criteria");
        filterCriteriaList = new JList();
        filterCriteriaList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        filterDescriptionLabel = new JLabel(
                "Filtering technique description and parameters");

        sortCheckBox = new JCheckBox("Sort");
        sortCheckBox.addItemListener(new
                                     PanelPostProcessing_sortCheckBox_itemAdapter(this));
        sortCriteriaLabel = new JLabel("Sorting Criteria");
        sortCriteriaList = new JList();
        sortCriteriaList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        sortDescriptionLabel = new JLabel(
                "Sorting technique description and parameters");

        this.add(filterCheckBox);
        this.add(filterCriteriaLabel);
        springLayout.putConstraint(SpringLayout.NORTH, filterCriteriaLabel, 5,
                                   SpringLayout.SOUTH, filterCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, filterCriteriaLabel, 20,
                                   SpringLayout.WEST, filterCheckBox);

        this.add(sortCheckBox);
        springLayout.putConstraint(SpringLayout.SOUTH, sortCheckBox, 0,
                                   SpringLayout.SOUTH, filterCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, sortCheckBox, 300,
                                   SpringLayout.EAST, filterCheckBox);

        this.add(sortCriteriaLabel);
        springLayout.putConstraint(SpringLayout.NORTH, sortCriteriaLabel, 5,
                                   SpringLayout.SOUTH, sortCheckBox);
        springLayout.putConstraint(SpringLayout.WEST, sortCriteriaLabel, 20,
                                   SpringLayout.WEST, sortCheckBox);

        applyButton = new JButton("Apply");
        applyButton.setPreferredSize(new Dimension(600, 50));
        applyButton.addActionListener(new
                                     PanelPostProcessing_applyButton_actionAdapter(this));
        this.add(applyButton);
        springLayout.putConstraint(SpringLayout.NORTH, applyButton, 160,
                                   SpringLayout.SOUTH, filterDescriptionLabel);
        springLayout.putConstraint(SpringLayout.WEST, applyButton, 30,
                                   SpringLayout.WEST, filterDescriptionLabel);

        ascendingButton = new JRadioButton("ascending", true);
        descendingButton = new JRadioButton("descending");
        group = new ButtonGroup();
        group.add(ascendingButton);
        group.add(descendingButton);
    }

    /**
     * Clears post-processing panel. Removes graphical components and sets
     * biclusters group and selection path to null.
     */
    public void clear() {
        this.biclustering = null;
        this.selectionPath = null;

        int components = this.getComponentCount();
        if (components > 5) {
            for (int i = 0; i < 6; i++) {
                this.remove(components - (1 + i));
            }
        }

        acvTextField.setEnabled(true);
        arvTextField.setEnabled(true);
        msrTextField.setEnabled(true);
        conditionsTextField.setEnabled(true);
        genesTextField.setEnabled(true);
        similarityTextField.setEnabled(true);
        sizeTextField.setEnabled(true);
        ascendingButton.setEnabled(true);
        descendingButton.setEnabled(true);
    }

    /**
     * Paints post-processing panel with post-processing options for a
     * specific biclusters group and tree selection path.
     *
     * @param biclustering <code>Biclustering</code> the biclusters group
     *        to post-process
     * @param selectionPath the <code>TreePath</code> to biclusters group object
     */
    public void paint(Biclustering biclustering, TreePath selectionPath) {
        this.biclustering = biclustering;
        this.selectionPath = selectionPath;
        if (biclustering instanceof CCC_Biclustering ||
            biclustering instanceof E_CCC_Biclustering) {
            paintDiscrete();
        } else {
            paintFloat();
        }
        this.setFilterCheckBoxSelected(true);
        this.setSortCheckBoxSelected(true);
    }

    /**
     * Paints post-processing object for biclusters groups with discrete data.
     */
    public void paintDiscrete() {
        if (biclustering == null) {
            return;
        }

        filterCriteriaList = new JList(this.getFilterCriteria());
        /*        if (biclustering instanceof E_CCC_Biclustering)
                    filterCriteriaList = new JList(filterCriteriaDiscrete);
                else
                    filterCriteriaList = new JList(filterCriteriaCCC);
         */
        filterCriteriaList.setSelectedIndex(0);
        filterCriteriaList.addListSelectionListener(new
                PanelPostProcessing_filterCriteriaList_listSelectionAdapter(this));

        sortCriteriaList = new JList(getSortCriteria());
        sortCriteriaList.setSelectedIndex(12);
        sortCriteriaList.addListSelectionListener(new
                                                  PanelPostProcessing_sortCriteriaList_listSelectionAdapter(this));

        sortPanel = new JPanel();
        sortPanel.setPreferredSize(new Dimension(500, 80));
        sortPanelLayout = new SpringLayout();
        sortPanel.setLayout(sortPanelLayout);

        JLabel label1 = new JLabel("sort biclusters by bicluster pattern");
        sortPanel.add(label1);
        JLabel label = new JLabel("p-value (ascending) with Markov chain order");
        sortPanel.add(label);
        sortPanelLayout.putConstraint(SpringLayout.NORTH, label, 5,
                                      SpringLayout.SOUTH, label1);
        sortPanelLayout.putConstraint(SpringLayout.WEST, label, 0,
                                      SpringLayout.WEST, label1);

        markovChainOrderSortTextField.setText("1");
        sortPanel.add(markovChainOrderSortTextField);
        sortPanelLayout.putConstraint(SpringLayout.SOUTH,
                                      markovChainOrderSortTextField, 0,
                                      SpringLayout.SOUTH, label);
        sortPanelLayout.putConstraint(SpringLayout.WEST,
                                      markovChainOrderSortTextField, 5,
                                      SpringLayout.EAST, label);

        this.paintBasic();
    }

    public void paintFloat() {
        filterCriteriaList = new JList(getFilterCriteria());
        filterCriteriaList.setSelectedIndex(0);
        filterCriteriaList.addListSelectionListener(new
                PanelPostProcessing_filterCriteriaList_listSelectionAdapter(this));

        sortCriteriaList = new JList(getSortCriteria());
        sortCriteriaList.setSelectedIndex(10);
        sortCriteriaList.addListSelectionListener(new
                                                  PanelPostProcessing_sortCriteriaList_listSelectionAdapter(this));

        sortPanel = new JPanel();
        sortPanel.setPreferredSize(new Dimension(500, 80));
        sortPanelLayout = new SpringLayout();
        sortPanel.setLayout(sortPanelLayout);

        JLabel label1 = new JLabel("sort biclusters by best GO Bonferroni");
        sortPanel.add(label1);
        JLabel label = new JLabel("corrected p-value (ascending)");
        sortPanel.add(label);
        sortPanelLayout.putConstraint(SpringLayout.NORTH, label, 5,
                                      SpringLayout.SOUTH, label1);
        sortPanelLayout.putConstraint(SpringLayout.WEST, label, 0,
                                      SpringLayout.WEST, label1);

        this.paintBasic();
    }

    public void paintBasic() {
        this.add(filterCriteriaList);
        springLayout.putConstraint(SpringLayout.NORTH, filterCriteriaList, 0,
                                   SpringLayout.NORTH, filterCriteriaLabel);
        springLayout.putConstraint(SpringLayout.WEST, filterCriteriaList, 5,
                                   SpringLayout.EAST, filterCriteriaLabel);

        this.add(filterDescriptionLabel);
        springLayout.putConstraint(SpringLayout.NORTH, filterDescriptionLabel,
                                   20,
                                   SpringLayout.SOUTH, filterCriteriaList);
        springLayout.putConstraint(SpringLayout.WEST, filterDescriptionLabel, 0,
                                   SpringLayout.WEST, filterCheckBox);

        this.add(sortCriteriaList);
        springLayout.putConstraint(SpringLayout.NORTH, sortCriteriaList, 0,
                                   SpringLayout.NORTH, sortCriteriaLabel);
        springLayout.putConstraint(SpringLayout.WEST, sortCriteriaList, 5,
                                   SpringLayout.EAST, sortCriteriaLabel);

        this.add(sortDescriptionLabel);
        springLayout.putConstraint(SpringLayout.NORTH, sortDescriptionLabel, 20,
                                   SpringLayout.SOUTH, sortCriteriaList);
        springLayout.putConstraint(SpringLayout.WEST, sortDescriptionLabel, 0,
                                   SpringLayout.WEST, sortCheckBox);

        filterPanel = new JPanel();
        filterPanel.setPreferredSize(new Dimension(300, 150));
        filterPanelLayout = new SpringLayout();
        filterPanel.setLayout(filterPanelLayout);
        filterPanel.setBorder(BorderFactory.createEmptyBorder(5, 0, 0, 0));

        this.add(filterPanel);
        springLayout.putConstraint(SpringLayout.NORTH, filterPanel, 5,
                                   SpringLayout.SOUTH, filterDescriptionLabel);
        springLayout.putConstraint(SpringLayout.WEST, filterPanel, 20,
                                   SpringLayout.WEST, filterDescriptionLabel);

        filterPanel.add(new JLabel("filter all trivial biclusters"));

        this.add(sortPanel);
        springLayout.putConstraint(SpringLayout.NORTH, sortPanel, 5,
                                   SpringLayout.SOUTH, sortDescriptionLabel);
        springLayout.putConstraint(SpringLayout.WEST, sortPanel, 20,
                                   SpringLayout.WEST, sortDescriptionLabel);

        this.setVisible(true);
    }

    public String[] getFilterCriteria() {
        if (biclustering == null) {
            return null;
        }
        if (biclustering instanceof BiclusteringInDiscretizedMatrix) {
            String[] criteria = {"trivial",
                                "genes",
                                "conditions",
                                "genes, conditions",
                                "trivial, ACV",
                                "trivial, ARV",
                                "trivial, ACV, ARV",
                                "trivial, MSR",
                                "trivial, size",
                                "trivial, size, ARV, ACV",
                                "conditions, genes, ARV, ACV",
                                "conditions, genes, ARV, ACV, MSR",
                                "overlapping genes",
                                "overlapping genes and conditions",
                                "constant biclusters",
                                "pattern (not significant)",
                                "pattern w/ columns (not significant)"};
            return criteria;
        }
        String[] criteria = {"trivial",
                            "genes",
                            "conditions",
                            "genes, conditions",
                            "trivial, ACV",
                            "trivial, ARV",
                            "trivial, ACV, ARV",
                            "trivial, MSR",
                            "trivial, size",
                            "trivial, size, ARV, ACV",
                            "conditions, genes, ARV, ACV",
                            "conditions, genes, ARV, ACV, MSR",
                            "overlapping genes",
                            "overlapping genes and conditions"};
        return criteria;
    }

    public String[] getSortCriteria() {
        if (biclustering == null) {
            return null;
        }
        if (biclustering instanceof CCC_Biclustering ||
            biclustering instanceof E_CCC_Biclustering) {
            String[] criteria = {"ACV",
                                "ARV",
                                "MSR",
                                "ARV-ACV",
                                "ARV-MSR",
                                "number of genes",
                                "number of conditions",
                                "size",
                                "variance",
                                "best GO p-value",
                                "best GO Bonferroni corrected p-value",
                                "bicluster pattern p-value",
                                "bicluster pattern w/ columns p-value"
            };
            return criteria;
        }
        String[] criteria = {"ACV",
                            "ARV",
                            "MSR",
                            "ARV-ACV",
                            "ARV-MSR",
                            "number of genes",
                            "number of conditions",
                            "size",
                            "variance",
                            "best GO p-value",
                            "best GO Bonferroni corrected p-value"};
        return criteria;
    }

    /**
     * Sets filter check box selected/deselected. Also other components are
     * set to enabled/disabled according to the state of filter check box.
     *
     * @param selected <code>boolean</code> <code>true</code> if check
     * box should be turn to selected state, <code>false</code> otherwise
     */
    public void setFilterCheckBoxSelected(boolean selected) {
        this.filterCheckBox.setSelected(selected);
        this.filterCriteriaLabel.setEnabled(selected);
        this.filterCriteriaList.setEnabled(selected);
        this.filterDescriptionLabel.setEnabled(selected);
        this.filterPanel.setEnabled(selected);
        for (Component c : this.filterPanel.getComponents()) {
            c.setEnabled(selected);
        }

        if (selected) {
            filterCriteriaList.setSelectedIndex(0);
        }
    }

    /**
     * Sets sort check box selected/deselected. Also other components are
     * set to enabled/disabled according to the state of sort check box.
     *
     * @param selected <code>boolean</code> <code>true</code> if check
     * box should be turn to selected state, <code>false</code> otherwise
     */
    public void setSortCheckBoxSelected(boolean selected) {
        this.sortCheckBox.setSelected(selected);
        this.sortCriteriaLabel.setEnabled(selected);
        this.sortCriteriaList.setEnabled(selected);
        this.sortDescriptionLabel.setEnabled(selected);
        this.sortPanel.setEnabled(selected);
        for (Component c : this.sortPanel.getComponents()) {
            c.setEnabled(selected);
        }

        if (selected) {
            if (sortCriteriaList.getModel().getSize() > 11) {
                sortCriteriaList.setSelectedIndex(12);
            } else {
                sortCriteriaList.setSelectedIndex(10);
            }
        }
    }

//############################################################################
//############################################################################
//############################################################################
//                    GETTERS AND SETTERS
//############################################################################
//############################################################################
//############################################################################

    /**
     * Returns the frame of the application.
     *
     * @return BiggestsFrame
     */
    public BiggestsFrame getMainFrame() {
        return mainFrame;
    }

    /**
     * Returns ACV's threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getACVTextField() {
        return acvTextField;
    }

    /**
     * Returns ARV's threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getARVTextField() {
        return arvTextField;
    }

    /**
     * Returns ascending order option radio button.
     *
     * @return JTextField
     */
    public JRadioButton getAscendingButton() {
        return ascendingButton;
    }

    /**
     * Returns conditions' number threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getConditionsTextField() {
        return conditionsTextField;
    }

    /**
     * Returns descending order option radio button.
     *
     * @return JTextField
     */
    public JRadioButton getDescendingButton() {
        return descendingButton;
    }

    /**
     * Returns enabling/disabling filtering option check box.
     *
     * @return JCheckBox
     */
    public JCheckBox getFilterCheckBox() {
        return filterCheckBox;
    }

    /**
     * Returns filtering criteria label.
     *
     * @return JLabel
     */
    public JLabel getFilterCriteriaLabel() {
        return filterCriteriaLabel;
    }

    /**
     * Returns filtering criteria list.
     *
     * @return JList
     */
    public JList getFilterCriteriaList() {
        return filterCriteriaList;
    }

    /**
     * Returns genes' number threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getGenesTextField() {
        return genesTextField;
    }

    /**
     * Returns Markov chain order input text field.
     *
     * @return JTextField
     */
    public JTextField getMarkovChainOrderFilterTextField() {
        return markovChainOrderFilterTextField;
    }

    /**
     * Returns Markov chain order input text field.
     *
     * @return JTextField
     */
    public JTextField getMarkovChainOrderSortTextField() {
        return markovChainOrderSortTextField;
    }


    /**
     * Returns MSR threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getMSRTextField() {
        return msrTextField;
    }

    /**
     * Returns enabling/disabling sorting option check box.
     *
     * @return JCheckBox
     */
    public JCheckBox getSortCheckBox() {
        return sortCheckBox;
    }

    /**
     * Returns sorting criteria list.
     *
     * @return JList
     */
    public JList getSortCriteriaList() {
        return sortCriteriaList;
    }

    /**
     * Returns the group of biclusters that is being post-processed.
     *
     * @return Biclustering
     */
    public Biclustering getBiclustering() {
        return biclustering;
    }

    /**
     * Returns the panel with filtering options.
     *
     * @return JPanel
     */
    public JPanel getFilterPanel() {
        return filterPanel;
    }

    /**
     * Returns the similarity threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getSimilarityTextField() {
        return similarityTextField;
    }

    /**
     * Rreturns the selected path in the dataset tree.
     *
     * @return TreePath
     */
    public TreePath getSelectionPath() {
        return selectionPath;
    }

    /**
     * Returns the size threshold input text field.
     *
     * @return JTextField
     */
    public JTextField getSizeTextField() {
        return sizeTextField;
    }

    /**
     * Returns the sorting criteria label.
     *
     * @return JLabel
     */
    public JLabel getSortCriteriaLabel() {
        return sortCriteriaLabel;
    }

    /**
     * Returns the panel with the sorting options.
     *
     * @return JPanel
     */
    public JPanel getSortPanel() {
        return sortPanel;
    }

    /**
     * Handles a selection change on filter criteria list. Displays
     * information about selected filter criteria and text fields for
     * the user to input filtering parameters.
     *
     * @param e the <code>ListSelectionEvent</code>
     */
    public void filterCriteriaList_valueChanged(ListSelectionEvent e) {
        if (filterPanel == null) {
            return;
        }
        filterPanel.removeAll();
        filterPanel.repaint();
        filterPanelLayout = new SpringLayout();
        filterPanel.setLayout(filterPanelLayout);
        int selectedIndex = this.filterCriteriaList.getSelectedIndex();
        switch (selectedIndex) {
        case 0:
            filterPanel.add(new JLabel("filter all trivial biclusters"));
            break;
        case 1:
            JLabel label1 = new JLabel("filter biclusters with less than");
            JLabel label2 = new JLabel("genes");
            filterPanel.add(label1);

            genesTextField.setText("3");
            filterPanel.add(genesTextField);
            filterPanelLayout.putConstraint(SpringLayout.WEST, genesTextField,
                                            5,
                                            SpringLayout.EAST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, genesTextField,
                                            0,
                                            SpringLayout.SOUTH, label1);

            filterPanel.add(label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 5,
                                            SpringLayout.EAST, genesTextField);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, label2, 0,
                                            SpringLayout.SOUTH, genesTextField);
            break;
        case 2:
            label1 = new JLabel("filter biclusters with less than");
            label2 = new JLabel("conditions");
            filterPanel.add(label1);
            conditionsTextField.setText("3");
            filterPanel.add(conditionsTextField);
            filterPanel.add(label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH,
                                            conditionsTextField, 0,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST,
                                            conditionsTextField, 5,
                                            SpringLayout.EAST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, label2, 0,
                                            SpringLayout.SOUTH,
                                            conditionsTextField);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 5,
                                            SpringLayout.EAST,
                                            conditionsTextField);
            break;
        case 3:
            label1 = new JLabel("filter biclusters with less than");
            label2 = new JLabel("genes");
            JLabel label3 = new JLabel("and less than");
            JLabel label4 = new JLabel("conditions");
            filterPanel.add(label1);
            genesTextField.setText("3");
            conditionsTextField.setText("3");
            filterPanel.add(genesTextField);
            filterPanel.add(label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, genesTextField,
                                            0,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, genesTextField,
                                            5,
                                            SpringLayout.EAST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, label2, 0,
                                            SpringLayout.SOUTH, genesTextField);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 5,
                                            SpringLayout.EAST, genesTextField);

            filterPanel.add(label3);
            filterPanel.add(conditionsTextField);
            filterPanel.add(label4);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH,
                                            conditionsTextField, 0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST,
                                            conditionsTextField, 5,
                                            SpringLayout.EAST, label3);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, label4, 0,
                                            SpringLayout.SOUTH,
                                            conditionsTextField);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label4, 5,
                                            SpringLayout.EAST,
                                            conditionsTextField);
            break;
        case 4:
            label1 = new JLabel("filter trivial biclusters and biclusters with");
            label2 = new JLabel("average column variance >");
            acvTextField.setText("0");
            filterPanel.add(label1);
            filterPanel.add(label2);
            filterPanel.add(acvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, acvTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, acvTextField, 5,
                                            SpringLayout.EAST, label2);
            break;
        case 5:
            label1 = new JLabel("filter trivial biclusters and biclusters with");
            label2 = new JLabel("average row variance <");
            arvTextField.setText("1");
            filterPanel.add(label1);
            filterPanel.add(label2);
            filterPanel.add(arvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, arvTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, arvTextField, 5,
                                            SpringLayout.EAST, label2);
            break;
        case 6:
            label1 = new JLabel("filter trivial biclusters and biclusters with");
            label2 = new JLabel("average column variance >");
            label3 = new JLabel("average row variance <");

            filterPanel.add(label1);
            filterPanel.add(label2);
            acvTextField.setText("0");
            filterPanel.add(acvTextField);

            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, acvTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, acvTextField, 5,
                                            SpringLayout.EAST, label2);

            filterPanel.add(label3);
            arvTextField.setText("1");
            filterPanel.add(arvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, arvTextField, 0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, arvTextField, 5,
                                            SpringLayout.EAST, label3);
            break;
        case 7:
            label1 = new JLabel("filter trivial biclusters and biclusters");
            label2 = new JLabel("with mean squared residue >");
            msrTextField.setText("0");
            filterPanel.add(label1);
            filterPanel.add(label2);
            filterPanel.add(msrTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, msrTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, msrTextField, 5,
                                            SpringLayout.EAST, label2);
            break;
        case 8:
            label1 = new JLabel("filter trivial biclusters and");
            label2 = new JLabel("biclusters with size <");
            sizeTextField.setText("6");
            filterPanel.add(label1);
            filterPanel.add(label2);
            filterPanel.add(sizeTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, sizeTextField,
                                            0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, sizeTextField, 5,
                                            SpringLayout.EAST, label2);
            break;
        case 9:
            label1 = new JLabel("filter trivial biclusters and biclusters with");
            label2 = new JLabel("size <");
            label3 = new JLabel("average column variance >");
            label4 = new JLabel("average row variance <");

            filterPanel.add(label1);
            filterPanel.add(label2);
            sizeTextField.setText("6");
            filterPanel.add(sizeTextField);

            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, sizeTextField,
                                            0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, sizeTextField, 5,
                                            SpringLayout.EAST, label2);

            filterPanel.add(label3);
            acvTextField.setText("0");
            filterPanel.add(acvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, acvTextField, 0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, acvTextField, 5,
                                            SpringLayout.EAST, label3);

            filterPanel.add(label4);
            arvTextField.setText("1");
            filterPanel.add(arvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label4, 5,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label4, 0,
                                            SpringLayout.WEST, label3);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, arvTextField, 0,
                                            SpringLayout.SOUTH, label4);
            filterPanelLayout.putConstraint(SpringLayout.WEST, arvTextField, 5,
                                            SpringLayout.EAST, label4);
            break;
        case 10:
            label1 = new JLabel("filter biclusters with a");
            label2 = new JLabel("number of conditions <");
            label3 = new JLabel("number of genes <");
            label4 = new JLabel("average column variance >");
            JLabel label5 = new JLabel("average row variance <");

            filterPanel.add(label1);
            filterPanel.add(label2);
            conditionsTextField.setText("3");
            filterPanel.add(conditionsTextField);

            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH,
                                            conditionsTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST,
                                            conditionsTextField, 5,
                                            SpringLayout.EAST, label2);

            filterPanel.add(label3);
            genesTextField.setText("3");
            filterPanel.add(genesTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, genesTextField,
                                            0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, genesTextField,
                                            5,
                                            SpringLayout.EAST, label3);

            filterPanel.add(label4);
            acvTextField.setText("0");
            filterPanel.add(acvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label4, 5,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label4, 0,
                                            SpringLayout.WEST, label3);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, acvTextField, 0,
                                            SpringLayout.SOUTH, label4);
            filterPanelLayout.putConstraint(SpringLayout.WEST, acvTextField, 5,
                                            SpringLayout.EAST, label4);

            filterPanel.add(label5);
            acvTextField.setText("0");
            filterPanel.add(arvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label5, 5,
                                            SpringLayout.SOUTH, label4);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label5, 0,
                                            SpringLayout.WEST, label4);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, arvTextField, 0,
                                            SpringLayout.SOUTH, label5);
            filterPanelLayout.putConstraint(SpringLayout.WEST, arvTextField, 5,
                                            SpringLayout.EAST, label5);
            break;
        case 11:
            label1 = new JLabel("filter biclusters with a");
            label2 = new JLabel("number of conditions <");
            label3 = new JLabel("number of genes <");
            label4 = new JLabel("average column variance >");
            label5 = new JLabel("average row variance <");
            JLabel label6 = new JLabel("mean squared residue >");

            filterPanel.add(label1);
            filterPanel.add(label2);
            conditionsTextField.setText("3");
            filterPanel.add(conditionsTextField);

            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH,
                                            conditionsTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST,
                                            conditionsTextField, 5,
                                            SpringLayout.EAST, label2);

            filterPanel.add(label3);
            genesTextField.setText("3");
            filterPanel.add(genesTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, genesTextField,
                                            0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, genesTextField,
                                            5,
                                            SpringLayout.EAST, label3);

            filterPanel.add(label4);
            acvTextField.setText("0");
            filterPanel.add(acvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label4, 5,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label4, 0,
                                            SpringLayout.WEST, label3);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, acvTextField, 0,
                                            SpringLayout.SOUTH, label4);
            filterPanelLayout.putConstraint(SpringLayout.WEST, acvTextField, 5,
                                            SpringLayout.EAST, label4);

            filterPanel.add(label5);
            arvTextField.setText("1");
            filterPanel.add(arvTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label5, 5,
                                            SpringLayout.SOUTH, label4);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label5, 0,
                                            SpringLayout.WEST, label4);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, arvTextField, 0,
                                            SpringLayout.SOUTH, label5);
            filterPanelLayout.putConstraint(SpringLayout.WEST, arvTextField, 5,
                                            SpringLayout.EAST, label5);

            filterPanel.add(label6);
            msrTextField.setText("0");
            filterPanel.add(msrTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label6, 5,
                                            SpringLayout.SOUTH, label5);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label6, 0,
                                            SpringLayout.WEST, label5);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, msrTextField, 0,
                                            SpringLayout.SOUTH, label6);
            filterPanelLayout.putConstraint(SpringLayout.WEST, msrTextField, 5,
                                            SpringLayout.EAST, label6);
            break;
        case 12:
            label1 = new JLabel("filter biclusters with a row similarity >");
            filterPanel.add(label1);
            similarityTextField.setText("25");
            filterPanel.add(similarityTextField);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH,
                                            similarityTextField, 0,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST,
                                            similarityTextField, 5,
                                            SpringLayout.EAST, label1);
            break;
        case 13:
            label1 = new JLabel("filter biclusters with");
            label2 = new JLabel("a row and column similarity >");
            filterPanel.add(label1);
            filterPanel.add(label2);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);

            similarityTextField.setText("25");
            filterPanel.add(similarityTextField);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH,
                                            similarityTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST,
                                            similarityTextField, 5,
                                            SpringLayout.EAST, label2);
            break;
        case 14:
            filterPanel.add(new JLabel("filter constant biclusters"));
            break;
        case 15:
            label1 = new JLabel("filter biclusters by pattern with");
            label2 = new JLabel("Markov chain order");
            label3 = new JLabel("p-value threshold");

            filterPanel.add(label1);
            filterPanel.add(label2);
            markovChainOrderFilterTextField.setText("1");
            filterPanel.add(markovChainOrderFilterTextField);

            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, markovChainOrderFilterTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, markovChainOrderFilterTextField, 5,
                                            SpringLayout.EAST, label2);

            filterPanel.add(label3);
            pValueTextField.setText("0.01");
            filterPanel.add(pValueTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, pValueTextField, 0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, pValueTextField, 5,
                                            SpringLayout.EAST, label3);
            break;
        case 16:
            label1 = new JLabel("filter biclusters by pattern w/ columns:");
            label2 = new JLabel("Markov chain order");
            label3 = new JLabel("p-value threshold");

            filterPanel.add(label1);
            filterPanel.add(label2);
            markovChainOrderFilterTextField.setText("1");
            filterPanel.add(markovChainOrderFilterTextField);

            filterPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                            SpringLayout.SOUTH, label1);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                            SpringLayout.WEST, label1);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, markovChainOrderFilterTextField, 0,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, markovChainOrderFilterTextField, 5,
                                            SpringLayout.EAST, label2);

            filterPanel.add(label3);
            pValueTextField.setText("0.01");
            filterPanel.add(pValueTextField);
            filterPanelLayout.putConstraint(SpringLayout.NORTH, label3, 5,
                                            SpringLayout.SOUTH, label2);
            filterPanelLayout.putConstraint(SpringLayout.WEST, label3, 0,
                                            SpringLayout.WEST, label2);
            filterPanelLayout.putConstraint(SpringLayout.SOUTH, pValueTextField, 0,
                                            SpringLayout.SOUTH, label3);
            filterPanelLayout.putConstraint(SpringLayout.WEST, pValueTextField, 5,
                                            SpringLayout.EAST, label3);
            break;
        }
        mainFrame.setVisible(true);
    }

    /**
     * Handles a selection change on sort criteria list. Displays
     * information about selected sort criteria and text fields for
     * the user to input sorting parameters.
     *
     * @param e the <code>ListSelectionEvent</code>
     */
    public void sortCriteriaList_valueChanged(ListSelectionEvent e) {
        if (sortPanel == null) {
            return;
        }

        sortPanel.removeAll();
        sortPanel.repaint();
        sortPanelLayout = new SpringLayout();
        sortPanel.setLayout(sortPanelLayout);

        ascendingButton.setSelected(true);
        boolean addButtons = true;

        int selectedIndex = this.sortCriteriaList.getSelectedIndex();
        JLabel label = new JLabel();
        switch (selectedIndex) {
        case 0:
            label.setText("sort biclusters by average column variance");
            sortPanel.add(label);
            break;
        case 1:
            label.setText("sort biclusters by average row variance");
            sortPanel.add(label);
            break;
        case 2:
            label.setText("sort biclusters by mean squared residue");
            sortPanel.add(label);
            break;
        case 3:
            label.setText("sort biclusters by ARV-ACV");
            sortPanel.add(label);
            break;
        case 4:
            label.setText("sort biclusters by ARV-MSR");
            sortPanel.add(label);
            break;
        case 5:
            label.setText("sort biclusters by number of genes");
            sortPanel.add(label);
            break;
        case 6:
            label.setText("sort biclusters by number of conditions");
            sortPanel.add(label);
            break;
        case 7:
            label.setText("sort biclusters by size");
            sortPanel.add(label);
            break;
        case 8:
            label.setText("sort biclusters by variance");
            sortPanel.add(label);
            break;
        case 9:
            label.setText("sort biclusters by best GO p-value (ascending)");
            sortPanel.add(label);
            addButtons = false;
            break;
        case 10:
            JLabel label1 = new JLabel("sort biclusters by best GO Bonferroni");
            sortPanel.add(label1);
            label.setText("corrected p-value (ascending)");
            sortPanel.add(label);
            sortPanelLayout.putConstraint(SpringLayout.NORTH, label, 5,
                                          SpringLayout.SOUTH, label1);
            sortPanelLayout.putConstraint(SpringLayout.WEST, label, 0,
                                          SpringLayout.WEST, label1);
            addButtons = false;
            break;
        case 11:
            label1 = new JLabel("sort biclusters by bicluster pattern");
            sortPanel.add(label1);
            label.setText("p-value (ascending) with Markov chain order");
            sortPanel.add(label);
            sortPanelLayout.putConstraint(SpringLayout.NORTH, label, 5,
                                          SpringLayout.SOUTH, label1);
            sortPanelLayout.putConstraint(SpringLayout.WEST, label, 0,
                                          SpringLayout.WEST, label1);

            markovChainOrderSortTextField.setText("1");
            sortPanel.add(markovChainOrderSortTextField);
            sortPanelLayout.putConstraint(SpringLayout.SOUTH,
                                          markovChainOrderSortTextField, 0,
                                          SpringLayout.SOUTH, label);
            sortPanelLayout.putConstraint(SpringLayout.WEST,
                                          markovChainOrderSortTextField, 5,
                                          SpringLayout.EAST, label);
            addButtons = false;
            break;
        case 12:
            label1 = new JLabel("sort biclusters by bicluster pattern");
            sortPanel.add(label1);
            label.setText("with columns p-value (ascending)");
            sortPanel.add(label);
            sortPanelLayout.putConstraint(SpringLayout.NORTH, label, 5,
                                          SpringLayout.SOUTH, label1);
            sortPanelLayout.putConstraint(SpringLayout.WEST, label, 0,
                                          SpringLayout.WEST, label1);
            JLabel label2 = new JLabel("with Markov chain order");
            sortPanel.add(label2);
            sortPanelLayout.putConstraint(SpringLayout.NORTH, label2, 5,
                                          SpringLayout.SOUTH, label);
            sortPanelLayout.putConstraint(SpringLayout.WEST, label2, 0,
                                          SpringLayout.WEST, label);

            markovChainOrderSortTextField.setText("1");
            sortPanel.add(markovChainOrderSortTextField);
            sortPanelLayout.putConstraint(SpringLayout.SOUTH,
                                          markovChainOrderSortTextField, 0,
                                          SpringLayout.SOUTH, label2);
            sortPanelLayout.putConstraint(SpringLayout.WEST,
                                          markovChainOrderSortTextField, 5,
                                          SpringLayout.EAST, label2);
            addButtons = false;
            break;
        }
        if (addButtons) {
            sortPanel.add(ascendingButton);
            sortPanelLayout.putConstraint(SpringLayout.NORTH, ascendingButton,
                                          5,
                                          SpringLayout.SOUTH, label);
            sortPanelLayout.putConstraint(SpringLayout.WEST, ascendingButton,
                                          30,
                                          SpringLayout.WEST, label);
            sortPanel.add(descendingButton);
            sortPanelLayout.putConstraint(SpringLayout.NORTH, descendingButton,
                                          5,
                                          SpringLayout.SOUTH, ascendingButton);
            sortPanelLayout.putConstraint(SpringLayout.WEST, descendingButton,
                                          0,
                                          SpringLayout.WEST, ascendingButton);
        }
        mainFrame.setVisible(true);
    }

    /**
     * Handles changes on filter check box selection.
     *
     * @param e <code>ItemEvent</code> the change on the state of the filter check box
     */
    public void filterCheckBox_itemStateChanged(ItemEvent e) {
        this.setFilterCheckBoxSelected(filterCheckBox.isSelected());
    }

    /**
     * Handles changes on sort check box selection.
     *
     * @param e <code>ItemEvent</code> the change on the state of the sort check box
     */
    public void sortCheckBox_itemStateChanged(ItemEvent e) {
        this.setSortCheckBoxSelected(sortCheckBox.isSelected());
    }

    /**
     * Reads a filtering/sorting parameter from its input text field box.
     * Valid parameter names:
     * PanelPostProcessing.PARAMETER_GENES
     * PanelPostProcessing.PARAMETER_CONDITIONS
     * PanelPostProcessing.PARAMETER_ACV
     * PanelPostProcessing.PARAMETER_ARV
     * PanelPostProcessing.PARAMETER_MSR
     * PanelPostProcessing.PARAMETER_SIZE
     * PanelPostProcessing.PARAMETER_SIMILARITY
     * PanelPostProcessing.PARAMETER_MARKOV
     *
     * @param parameterToRead <code>String</code> the name of the
     *        parameter to read
     * @return String the parameter read from the input text field
     * @throws Exception if parameter could not be read
     * @see PanelBasicIOValidation#readParameter(String parameterToRead)
     */
    public String readParameter(String parameterToRead) throws Exception {
        String parameter = null;
        if (parameterToRead.equals(this.PARAMETER_GENES)) {
            parameter = this.genesTextField.getText();
        } else
        if (parameterToRead.equals(this.PARAMETER_CONDITIONS)) {
            parameter = this.conditionsTextField.getText();
        } else
        if (parameterToRead.equals(this.PARAMETER_ACV)) {
            parameter = this.acvTextField.getText();
        } else
        if (parameterToRead.equals(this.PARAMETER_ARV)) {
            parameter = this.arvTextField.getText();
        } else
        if (parameterToRead.equals(this.PARAMETER_MSR)) {
            parameter = this.msrTextField.getText();
        } else
        if (parameterToRead.equals(this.PARAMETER_SIZE)) {
            parameter = this.sizeTextField.getText();
        } else
        if (parameterToRead.equals(this.PARAMETER_SIMILARITY)) {
            parameter = this.similarityTextField.getText();
        }
        else
        if (parameterToRead.equals(this.PARAMETER_MARKOV_FILTER)) {
            parameter = this.markovChainOrderFilterTextField.getText();
        }
        else
        if (parameterToRead.equals(this.PARAMETER_MARKOV_SORT)) {
            parameter = this.markovChainOrderSortTextField.getText();
        }
        else
        if (parameterToRead.equals(this.PARAMETER_PVALUE)) {
            parameter = this.pValueTextField.getText();
        }
        if ((parameter == null) ||
            (parameter.equals(""))) {
            JOptionPane.showMessageDialog(this, "Please type the "
                                          + parameterToRead + ".",
                                          "Type " + parameterToRead,
                                          JOptionPane.INFORMATION_MESSAGE);
            throw (new Exception());
        }
        return parameter;
    }

    /**
     * Applies the post-processing criteria to the current biclusters group.
     *
     * @param e <code>ActionEvent</code> the click on apply button
     */
    public void applyButton_actionPerformed(ActionEvent e) {
        boolean filter = this.filterCheckBox.isSelected();
        boolean sort = this.sortCheckBox.isSelected();
        if (!filter && !sort) {
            return;
        }

        ThreadPostProcessing postProcessingThread = new ThreadPostProcessing(
                biclustering,
                selectionPath,
                this);
        postProcessingThread.start();
    }

    /**
     * Copies the function result files for a post-processed
     * biclusters' group.
     *
     * @param bic <code>Biclustering</code> the post-processed
     *        group of biclusters to add to the dataset tree
     * @param infoStrings the <code>ArrayList</code> with the set of
     *        information strings
     * @param infoStyles the <code>ArrayList</code> with the set of
     *        formatting styles for <code>infoStrings</code>
     * @param selectionPath the dataset tree selected <code>TreePath</code>
     */
    public void copyFunctionResultFilesForPostProcessedBiclusterGroup(
            TreePath selectionPath) throws Exception {
        int newPostProcessedID = mainFrame.getTree().generateNewNodeInfoId(
            NodeInfo.POSTPROCESSED_GROUP, this.selectionPath, false);
        String pathToPostProcessedFunctionResults = mainFrame.
                                                    getPathToFunctionResults()
                                                    + File.separator +
                                                    selectionPath.getPath()[1] +
                                                    File.separator +
                                                    "Post-Processed Group " +
                                                    newPostProcessedID;
        File directoryPostProcessedFunctionResults = new File(
            pathToPostProcessedFunctionResults);
        directoryPostProcessedFunctionResults.mkdir();

        NodeInfo biclusteringNodeInfo = (NodeInfo)
                                        ((DefaultMutableTreeNode)
                                         selectionPath.getLastPathComponent())
                                        .getUserObject();

        String pathToBiclusteringFunctionResults =
            mainFrame.getPathToFunctionResults() +
            File.separator + selectionPath.getPath()[1].toString() +
            File.separator + biclusteringNodeInfo.toString();

        File directoryBiclusteringFunctionResults = new File(
            pathToBiclusteringFunctionResults);

        if (directoryBiclusteringFunctionResults.listFiles() != null) {
            for (File file :
                 (File[]) directoryBiclusteringFunctionResults.listFiles()) {
                BiggestsIO.copy(file,
                                new File(directoryPostProcessedFunctionResults.
                                         getAbsoluteFile() + File.separator +
                                         file.getName()));
            }
        }
    }

    /**
     * Adds a post-processed biclustering to the dataset tree.
     *
     * @param bic <code>Biclustering</code> the post-processed
     *        bicluster group to add to the dataset tree
     * @param infoStrings <code>ArrayList</code> the set of information strings
     * @param infoStyles <code>ArrayList</code> the set of styles for the
     *        information strings
     * @param selectionPath the selected <code>TreePath</code> in the
     *        dataset tree, which contains the biclusters' group from which
     *        this post-processed group of biclusters was derived
     * @throws Exception
     */
    public void addPostProcessedBiclusterGroupToTree(Biclustering bic,
            ArrayList<String> infoStrings, ArrayList<String> infoStyles,
            TreePath selectionPath) throws Exception {

        infoStrings.add(0, "Number of biclusters: ");
        infoStyles.add(0, "bold");
        infoStrings.add(1, bic.getNumberOfBiclusters() + "\n\n");
        infoStyles.add(1, "regular");
        infoStrings.add(2, "Datatype: ");
        infoStyles.add(2, "bold");
        if (bic instanceof CCC_Biclustering ||
            bic instanceof E_CCC_Biclustering) {
                infoStrings.add(3, "discrete symbols\n");
                infoStyles.add(3, "regular");
            } else {
                infoStrings.add(3, "floating point values\n");
                infoStyles.add(3, "regular");
            }

            infoStrings.add(4, "Organism: ");
            infoStyles.add(4, "bold");
            String organism = "";
            if (bic.getMatrix() instanceof AbstractExpressionMatrix) {
                organism = ((AbstractExpressionMatrix) bic.getMatrix()).
                           getOrganism();
            } else {
                organism = ((DiscretizedExpressionMatrix) bic.getMatrix()).
                           getOriginalExpressionMatrix().getOrganism();
            }
            infoStrings.add(5, organism + "\n\n");
            infoStyles.add(5, "regular");
/*            if (bic.getGoTermsComputed()) {
                infoStrings.add(6, "Function analysis: ");
                infoStyles.add(6, "bold");
                infoStrings.add(7, "p-value threshold - " +
                                biclustering.getGO_pValues_userThreshold() +
                                "\n");
                infoStyles.add(7, "regular");
            }
 */

            NodeInfo biclusteringNodeInfo = (NodeInfo)
                                            ((DefaultMutableTreeNode)
                                             selectionPath.getLastPathComponent())
                                            .getUserObject();

            NodeInfo postProcessedBiclusteringNodeInfo = new NodeInfo(
            bic,
            NodeInfo.POSTPROCESSED_GROUP,
            selectionPath,
            mainFrame.getTree(),
            biclusteringNodeInfo,
            infoStrings,
            infoStyles);

            TreePath newPath = mainFrame.getTree().addNewNode(
            postProcessedBiclusteringNodeInfo,
            selectionPath);
            mainFrame.getPanelAnalyzingFunctionTable().putGOFrontEnd(newPath,
                    mainFrame.getPanelAnalyzingFunctionTable().
                    getGOFrontEndForSelectionPath(selectionPath));
            mainFrame.getTree().setSelectionPath(newPath);
        }

        /**
         * Displays a message when some problem occurs while performing
         * filtering operations.
         *
         * @param sort the <code>boolean</code> value that states if sort
         *        option was enabled (<code>true</code>) or not (<code>false</code>)
         */
        public void displayFilteringErrorMessage(boolean sort) {
            String s = "";
            if (sort) {
                s += "You may want to deselect filtering option " +
                        "to perform only sorting operation.";
        }
        JOptionPane.showMessageDialog(this,
                                      "Filtering failed.\n" +
                                      "Therefore, biclusters group will not " +
                                      "be added to the dataset tree.\n" + s,
                                      "Filtering failed",
                                      JOptionPane.INFORMATION_MESSAGE);
    }
}


/**
 * <p>Description: Action adapter for apply button.</p>
 */
class PanelPostProcessing_applyButton_actionAdapter implements ActionListener {
    private PanelPostProcessing adaptee;
    PanelPostProcessing_applyButton_actionAdapter(PanelPostProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void actionPerformed(ActionEvent e) {
        adaptee.applyButton_actionPerformed(e);
    }
}


/**
 * <p>Description: Item adapter for sort check box.</p>
 */
class PanelPostProcessing_sortCheckBox_itemAdapter implements ItemListener {
    private PanelPostProcessing adaptee;
    PanelPostProcessing_sortCheckBox_itemAdapter(PanelPostProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.sortCheckBox_itemStateChanged(e);
    }
}


/**
 * <p>Description: Item adapter for filter check box.</p>
 */
class PanelPostProcessing_filterCheckBox_itemAdapter implements ItemListener {
    private PanelPostProcessing adaptee;
    PanelPostProcessing_filterCheckBox_itemAdapter(PanelPostProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.filterCheckBox_itemStateChanged(e);
    }
}


/**
 * <p>Description: List selection adapter for sort criteria list.</p>
 */
class PanelPostProcessing_sortCriteriaList_listSelectionAdapter implements
        ListSelectionListener {
    private PanelPostProcessing adaptee;
    PanelPostProcessing_sortCriteriaList_listSelectionAdapter(
            PanelPostProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void valueChanged(ListSelectionEvent e) {
        adaptee.sortCriteriaList_valueChanged(e);
    }
}


/**
 * <p>Description: List selection adapter for filter criteria list.</p>
 */
class PanelPostProcessing_filterCriteriaList_listSelectionAdapter implements
        ListSelectionListener {
    private PanelPostProcessing adaptee;
    PanelPostProcessing_filterCriteriaList_listSelectionAdapter(
            PanelPostProcessing adaptee) {
        this.adaptee = adaptee;
    }

    public void valueChanged(ListSelectionEvent e) {
        adaptee.filterCriteriaList_valueChanged(e);
    }
}
