package biggests.gui;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.tree.*;
import biggests.utils.*;
import de.schlichtherle.io.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;
import java.util.Hashtable;
import smadeira.utils.BiologicalRelevanceSummary;
import java.io.PrintWriter;

/**
 * <p>Title: Function analysis thread.</p>
 *
 * <p>Description: Thread to compute function analysis
 *                (long time consuming task) while displaying a progress
 *                bar dialog to the user.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana Gonalves
 */
public class ThreadFunctionAnalysis extends Thread {
    /**
     * Node object: it can be a matrix or a bicluster group.
     */
    private NodeObjectInterface nodeObject;

    private TreePath selectionPath;

    /**
     * The dialog which displays the progress bar.
     */
    private ProgressBarFrame progressFrame;

    /**
     * The main frame of the application.
     */
    private BiggestsFrame mainFrame;

    /**
     * Gene ontology files.
     * goFiles[0] - gene ontology file
     * goFiles[1] - gene association file
     */
    private File[] goFiles;

    private GOFrontEnd goFrontEnd;

    /**
     * Directory (or file, if nodeObject is an instance of one of
     * <code>IMatrix</code>'s subclasses) to store files resulting
     * from biological relevance computation.
     */
    private File functionOutput;

    /**
     * Minimum corrected pvalue to identify relevant biological functions.
     */
    private float pvalue;

    /**
     * Creates a new <code>ThreadFunctionAnalysis</code>.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param biclustering <code>Biclustering</code> object to perform
     *        function analysis on to
     * @param selectionPath the <code>TreePath</code> which corresponds to
     *        <code>biclustering</code> place in the dataset tree
     * @param goFiles <code>File[]</code> gene ontology files
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param pvalue float
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  Biclustering biclustering,
                                  TreePath selectionPath,
                                  File[] goFiles,
                                  File directoryFunctionResults,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");

        if (!this.areGOFilesValid(goFiles)) {
            throw new Exception("Invalid gene ontology files.");
        }
        this.nodeObject = biclustering;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFiles = goFiles;
        this.functionOutput = directoryFunctionResults;
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();
    }

    /**
     * Creates a new <code>ThreadFunctionAnalysis</code>.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param biclustering <code>Biclustering</code> object to perform
     *        function analysis on to
     * @param selectionPath the <code>TreePath</code> which corresponds to
     *        <code>biclustering</code> place in the dataset tree
     * @param go the <code>GOFrontEnd</code> for the given
     *        <code>biclustering</code>
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param pvalue <code>float</code> the p-value threshold for
     *        significant terms
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  Biclustering biclustering,
                                  TreePath selectionPath,
                                  GOFrontEnd go,
                                  File directoryFunctionResults,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");

        this.nodeObject = biclustering;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFrontEnd = go;
        this.functionOutput = directoryFunctionResults;
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();
    }


    /**
     * Creates a new <code>ThreadFunctionAnalysis</code>.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param biclustering <code>Biclustering</code> object to perform
     *        function analysis on to
     * @param selectionPath the <code>TreePath</code> which corresponds to
     *        <code>biclustering</code> place in the dataset tree
     * @param pathToGeneOntologyFile <code>String</code> the path to gene
     *        ontology file
     * @param pathToGeneAssociationFile <code>String</code> the path to
     *        gene association file
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param pvalue float
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  Biclustering biclustering,
                                  TreePath selectionPath,
                                  String pathToGeneOntologyFile,
                                  String pathToGeneAssociationFile,
                                  File directoryFunctionResults,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");
        if (pathToGeneAssociationFile == null ||
            pathToGeneAssociationFile == null) {
            throw new Exception("Invalid gene ontology files.");
        }
        File[] goFiles = {new File(pathToGeneOntologyFile),
                         new File(pathToGeneAssociationFile)};

        if (!this.areGOFilesValid(goFiles)) {
            throw new Exception("Invalid gene ontology files.");
        }
        this.nodeObject = biclustering;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFiles = goFiles;
        this.functionOutput = directoryFunctionResults;
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();
    }

    /**
     * Creates a new Function Analysis Thread to perform function
     * analysis over a matrix.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param matrix the <code>IMatrix</code> over which function
     *               analysis will be performed
     * @param selectionPath the <code>TreePath</code> which corresponds to
     *        <code>biclustering</code> place in the dataset tree
     * @param goFiles <code>File[]</code> gene ontology files
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param filenameDOT String the name for the resulting DOT file
     *        (graph of significant terms)
     * @param pvalue <code>float</code> the p-value threshold for
     *        significant terms
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  IMatrix matrix,
                                  TreePath selectionPath,
                                  File[] goFiles,
                                  File directoryFunctionResults,
                                  String filenameDOT,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");
        if (!this.areGOFilesValid(goFiles)) {
            throw new Exception("Invalid gene ontology files.");
        }
        this.nodeObject = matrix;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFiles = goFiles;
        this.functionOutput = new File(directoryFunctionResults.
                                       getAbsolutePath() + File.separator +
                                       filenameDOT);
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();

    }

    /**
     * Creates a new Function Analysis Thread to perform function
     * analysis over a matrix.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param matrix the <code>IMatrix</code> over which function
     *               analysis will be performed
     * @param selectionPath the <code>TreePath</code> which corresponds to
     *        <code>biclustering</code> place in the dataset tree
     * @param pathToGeneOntologyFile <code>String</code> the path to gene
     *        ontology file
     * @param pathToGeneAssociationFile <code>String</code> the path to
     *        gene association file
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param filenameDOT String the name for the resulting DOT file
     *        (graph of significant terms)
     * @param pvalue <code>float</code> the p-value threshold for
     *        significant terms
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  IMatrix matrix,
                                  TreePath selectionPath,
                                  String pathToGeneOntologyFile,
                                  String pathToGeneAssociationFile,
                                  File directoryFunctionResults,
                                  String filenameDOT,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");

        if (pathToGeneAssociationFile == null ||
            pathToGeneAssociationFile == null) {
            throw new Exception("Invalid gene ontology files.");
        }
        File[] goFiles = {new File(pathToGeneOntologyFile),
                         new File(pathToGeneAssociationFile)};

        if (!this.areGOFilesValid(goFiles)) {
            return;
        }
        this.nodeObject = matrix;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFiles = goFiles;
        this.functionOutput = new File(directoryFunctionResults.
                                       getAbsolutePath() + File.separator +
                                       filenameDOT);
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();

    }

    /**
     * Creates a new Function Analysis Thread to perform function
     * analysis over a matrix.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param bicluster the <code>Bicluster</code> over which function
     *        analysis will be performed
     * @param selectionPath the selected <code>TreePath</code> in the
     *        dataset tree (corresponding to the <code>bicluster</code> node)
     * @param goFiles <code>File[]</code> gene ontology files
     * @param directoryBiclusteringFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param pvalue <code>float</code> the p-value threshold for
     *        significant terms
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  Bicluster bicluster,
                                  TreePath selectionPath,
                                  File[] goFiles,
                                  File directoryBiclusteringFunctionResults,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");
        if (!this.areGOFilesValid(goFiles)) {
            throw new Exception("Invalid gene ontology files.");
        }
        this.nodeObject = bicluster;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFiles = goFiles;
        this.functionOutput = directoryBiclusteringFunctionResults;
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();

    }

    /**
     * Creates a new Function Analysis Thread to perform function
     * analysis over a matrix.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param bicluster the <code>Bicluster</code> on which function
     *        analysis has to be performed
     * @param selectionPath the selected <code>TreePath</code> in the
     *        dataset tree (corresponding to the <code>bicluster</code> node
     * @param pathToGeneOntologyFile String the path to gene ontology file
     *        (OBO file)
     * @param pathToGeneAssociationFile String the path to gene
     *        association file (organism specific file)
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param pvalue <code>float</code> the p-value threshold for
     *        significant terms
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  Bicluster bicluster,
                                  TreePath selectionPath,
                                  String pathToGeneOntologyFile,
                                  String pathToGeneAssociationFile,
                                  String directoryFunctionResults,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");

        if (pathToGeneAssociationFile == null ||
            pathToGeneAssociationFile == null) {
            throw new Exception("Invalid gene ontology files.");
        }
        File[] goFiles = {new File(pathToGeneOntologyFile),
                         new File(pathToGeneAssociationFile)};

        if (!this.areGOFilesValid(goFiles)) {
            throw new Exception("Invalid gene ontology files.");
        }
        this.nodeObject = bicluster;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFiles = goFiles;
        this.functionOutput = new File(directoryFunctionResults);
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();
    }

    /**
     * Creates a new <code>ThreadFunctionAnalysis</code>.
     *
     * @param frame the main <code>BiggestsFrame</code>
     * @param bicluster <code>Bicluster</code> object to perform
     *        function analysis on to
     * @param selectionPath the <code>TreePath</code> which corresponds to
     *        <code>bicluster</code>'s place in the dataset tree
     * @param go the <code>GOFrontEnd</code> for the given
     *        <code>bicluster</code>
     * @param directoryFunctionResults <code>File</code> directory
     *        to store the files resulting from function analysis
     * @param pvalue <code>float</code> the p-value threshold for
     *        significant terms
     * @throws Exception
     */
    public ThreadFunctionAnalysis(BiggestsFrame frame,
                                  Bicluster bicluster,
                                  TreePath selectionPath,
                                  GOFrontEnd go,
                                  File directoryFunctionResults,
                                  float pvalue) throws Exception {
        super("Function Analysis Thread");
        this.nodeObject = bicluster;
        this.selectionPath = selectionPath;
        mainFrame = frame;
        this.goFrontEnd = go;
        this.functionOutput = directoryFunctionResults;
        this.pvalue = pvalue;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();
    }


    /**
     * Overrides <code>Thread</code>'s  <code>run</code> method. Performs the
     * actual task of this working thread (performing function analysis on
     * a given biclusters group).
     *
     * @see Thread#run()
     */
    public void run() {
        progressFrame.setString("performing function analysis...");
        try {
            NodeInfo node = (NodeInfo) ((DefaultMutableTreeNode) selectionPath.
                                        getLastPathComponent()).getUserObject();
            PanelAnalyzingFunctionTable functionsPanel =
                    mainFrame.getPanelAnalyzingFunctionTable();
            if (nodeObject instanceof Biclustering) {
                if (goFrontEnd == null) {
                    goFrontEnd = functionsPanel.calculateGOFrontEnd(
                            (Biclustering)nodeObject,
                            goFiles[0].getAbsolutePath(),
                            goFiles[1].getAbsolutePath());
                    functionsPanel.putGOFrontEnd(selectionPath, goFrontEnd);
                    if (goFrontEnd == null) {
                        return;
                    }
                }

                Hashtable<Integer, BiologicalRelevanceSummary>
                        biologicalRelevanceSummary = ((
                                Biclustering)nodeObject).
                        computeBiologicalRelevanceUsingOntologizer(
                                goFrontEnd, pvalue);

                mainFrame.getPanelAnalyzingMatrixColors().clear();
                mainFrame.getPanelAnalyzingMatrixColors().paint(
                        node, selectionPath);
                mainFrame.getPanelAnalyzingMatrixValues().clear();
                mainFrame.getPanelAnalyzingMatrixValues().paint(node,
                        selectionPath);
                if (nodeObject instanceof BiclusteringInDiscretizedMatrix) {
                    mainFrame.getPanelAnalyzingMatrixSymbols().clear();
                    mainFrame.getPanelAnalyzingMatrixSymbols().paint(node,
                            selectionPath);
                }

                progressFrame.dispose();
                functionsPanel.
                        repaintPanelAnalyzingFunctionTable(
                                biologicalRelevanceSummary);
            } else {
                if (nodeObject instanceof AbstractExpressionMatrix) {
                    if (goFrontEnd == null) {
                        goFrontEnd = functionsPanel.calculateGOFrontEnd(
                                (AbstractExpressionMatrix)nodeObject,
                                goFiles[0].getAbsolutePath(),
                                goFiles[1].getAbsolutePath());
                        functionsPanel.putGOFrontEnd(selectionPath, goFrontEnd);
                        if (goFrontEnd == null) {
                            return;
                        }
                    }
                    ((AbstractExpressionMatrix)nodeObject).
                            computeGeneOntologyTerms(goFrontEnd, false);

                    mainFrame.getPanelAnalyzingMatrixColors().clear();
                    mainFrame.getPanelAnalyzingMatrixColors().paint(
                            node, selectionPath);
                    mainFrame.getPanelAnalyzingMatrixValues().clear();
                    mainFrame.getPanelAnalyzingMatrixValues().paint(node,
                            selectionPath);

                    progressFrame.dispose();
                    functionsPanel.
                        repaintPanelAnalyzingFunctionTable();

                }
                else
                if (nodeObject instanceof DiscretizedExpressionMatrix) {
                    // discretized expression matrix
                    if (goFrontEnd == null) {
                        goFrontEnd = functionsPanel.calculateGOFrontEnd(
                                (DiscretizedExpressionMatrix)nodeObject,
                                goFiles[0].getAbsolutePath(),
                                goFiles[1].getAbsolutePath());
                        functionsPanel.putGOFrontEnd(selectionPath, goFrontEnd);
                        if (goFrontEnd == null) {
                            return;
                        }
                    }
                    ((DiscretizedExpressionMatrix) nodeObject).
                            getOriginalExpressionMatrix().
                            computeGeneOntologyTerms(goFrontEnd, false);
                    mainFrame.getPanelAnalyzingMatrixValues().clear();
                    mainFrame.getPanelAnalyzingMatrixValues().paint(node,
                            selectionPath);
                    mainFrame.getPanelAnalyzingMatrixColors().clear();
                    mainFrame.getPanelAnalyzingMatrixColors().paint(
                            node, selectionPath);
                    mainFrame.getPanelAnalyzingMatrixSymbols().clear();
                    mainFrame.getPanelAnalyzingMatrixSymbols().paint(node,
                            selectionPath);

                    progressFrame.dispose();
                    functionsPanel.
                            repaintPanelAnalyzingFunctionTable();
                }
                else
                if (nodeObject instanceof Bicluster) {
                    if (goFrontEnd == null) {
                        goFrontEnd = functionsPanel.calculateGOFrontEnd(
                                (Bicluster)nodeObject,
                                goFiles[0].getAbsolutePath(),
                                goFiles[1].getAbsolutePath());
                        functionsPanel.putGOFrontEnd(selectionPath, goFrontEnd);
                        if (goFrontEnd == null) {
                            return;
                        }
                    }

                    Bicluster b = (Bicluster)nodeObject;
                    GeneSetResult goProperties =
                            b.
                            computeGeneOntologyTerms(goFrontEnd, false);
                    boolean hasGraph = false;
                    if (goProperties != null &&
                        b.computeSignificant_GO_BonferroniCorrected_pValues_UserThreshold(goProperties, pvalue) > 0) {
                        hasGraph = true;
                        b.printGraphWithGeneOntologyTerms_ToFile(
                                goFrontEnd, functionOutput.getAbsolutePath(),
                                "graph-BICLUSTER_" + b.getID() + ".dot", pvalue);
                    }

                    mainFrame.getPanelAnalyzingMatrixColors().clear();
                    mainFrame.getPanelAnalyzingMatrixColors().paint(
                            node, selectionPath);
                    mainFrame.getPanelAnalyzingMatrixValues().clear();
                    mainFrame.getPanelAnalyzingMatrixValues().paint(node,
                            selectionPath);
                    if (b instanceof CCC_Bicluster) {
                        mainFrame.getPanelAnalyzingMatrixSymbols().clear();
                        mainFrame.getPanelAnalyzingMatrixSymbols().paint(node,
                                selectionPath);
                    }

                    progressFrame.dispose();
                    functionsPanel.
                            repaintPanelAnalyzingFunctionTable(
                                    goProperties,
                                    pvalue);
                    if (hasGraph) {
                        mainFrame.
                                getPanelAnalyzingFunctionSignificanceGraph().
                                clear();
                        mainFrame.
                                getPanelAnalyzingFunctionSignificanceGraph().
                                paint((Bicluster) nodeObject, selectionPath);
                    }
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            progressFrame.dispose();
            try {
                PrintWriter pw = new PrintWriter(new FileOutputStream(
                        "test.txt"));
                e.printStackTrace(pw);
                pw.close();
            }
            catch (Exception ex) {
                e.printStackTrace();
            }
            JOptionPane.showMessageDialog(mainFrame, "Could not compute " +
                                          "biological relevance using " +
                                          "Ontologizer" + e.getCause().getClass(),
                                          "Couldn't compute functions" +
                                          "\n",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        } catch (StackOverflowError e) {
            e.printStackTrace();
            progressFrame.dispose();
            JOptionPane.showMessageDialog(mainFrame, "Could not compute " +
                                          "biological relevance using " +
                                          "Ontologizer: stack overflow.",
                                          "Couldn't compute functions",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        } catch (OutOfMemoryError e) {
            e.printStackTrace();
            progressFrame.dispose();
            JOptionPane.showMessageDialog(mainFrame, "Could not compute " +
                                          "biological relevance using " +
                                          "Ontologizer: out of memory.",
                                          "Couldn't compute functions",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }
    }

    /**
     * Checks if the given <code>goFiles</code> are valid files
     * for function analysis.
     *
     * @param goFiles the set of <code>File[]</code> to check
     * @return boolean <code>true</code> if files are valid for
     *         function analysis, <code>false</code> otherwise
     */
    private boolean areGOFilesValid(File[] goFiles) {
        if (goFiles.length != 2 ||
            goFiles[0] == null ||
            goFiles[1] == null ||
            !goFiles[0].exists() ||
            !goFiles[1].exists() ||
            !goFiles[0].getName().equals("gene_ontology.obo") ||
            !goFiles[1].getName().startsWith("gene_association")) {
            return false;
        }
        return true;
    }
}
