package biggests.tables;

import java.awt.*;
import java.util.*;

import javax.swing.*;
import javax.swing.table.*;

/**
 * <p>Title: Preview Table</p>
 *
 * <p>Description:  A preview table displays: names of genes in the
 *                  first column, names of conditions in the header
 *                  (first row) and gene expression values in remaining
 *                  cells. More widely, the content can be:
 *                  - first few rows: conditions' names (more widely,
 *                  <code>String</code>s)<br/>
 *                  - first few columns: genes' names (more widely,
 *                  <code>String</code>s)<br/>
 *                  - all remaining cells: gene expression values (more widely,
 *                  <code>Float</code>s)<br/>
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana Gonalves
 */
/*
 * Sample:
 * ---------------------------------------------------------------------<br/>
 * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
 * |-------------------------------------------------------------------|<br/>
 * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
 * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
 * | Gene3       | .......... | .......... | .......... | ............ |<br/>
 * | Gene4       | .......... | .......... | .......... | ............ |<br/>
 * ---------------------------------------------------------------------
 */
public class TablePreview extends JTable {
    /**
     * Indexes of rows which contain names (possibly conditions' names).
     */
    private ArrayList<Integer> rowsWithNames;

    /**
     * Indexes of columns which contain names (possibly genes' names).
     */
    private ArrayList<Integer> columnsWithNames;

    /**
     * Creates a new preview table with the given row data and column names.
     *
     * @param rowData the <code>Object[][]</code> row data (for all table
     *                cells except the ones from the table header)
     * @param columnNames the <code>Object[]</code> column names (for
     *                the table header)
     * @param rowsWithNames the <code>ArrayList<Integer></code> with the
     *        indexes of the rows that contain only names (of conditions)
     * @param columnsWithNames  the <code>ArrayList<Integer></code> with the
     *        indexes of the columns that contain only names (of genes)
     * @see #TablePreview(TableModel tableModel, ArrayList<Integer> rowsWithNames, ArrayList<Integer> columnsWithNames)
     * @see javax.swing.table.DefaultTableModel#DefaultTableModel(Object[][] rowData, Object[] columnNames)
     */
    public TablePreview(Object[][] rowData, Object[] columnNames,
                        ArrayList<Integer> rowsWithNames,
            ArrayList<Integer> columnsWithNames) {
        /**
         * create a new JTable with a table sorter model, which incorporates
         * the specified row data and columns' names
         */
        this(new DefaultTableModel(rowData, columnNames), rowsWithNames,
             columnsWithNames);
    }

    /**
     * Creates a new preview table with the given row data and column names.
     *
     * @param tableModel the <code>TableModel</code> for the table
     * @param rowsWithNames the <code>ArrayList<Integer></code> with the
     *        indexes of the rows that contain only names (of conditions)
     * @param columnsWithNames  the <code>ArrayList<Integer></code> with the
     *        indexes of the columns that contain only names (of genes)
     * @see javax.swing.JTable#JTable(TableModel dm)
     */
    protected TablePreview(TableModel tableModel,
                           ArrayList<Integer> rowsWithNames,
            ArrayList<Integer> columnsWithNames) {

        /**
         * create a new JTable object with a table sorter model derived from the
         * given table model
         */
        super(tableModel);

        this.getTableHeader().setReorderingAllowed(false);

        /**
         * set autoresize (of columns' width) off
         */
        this.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

        /**
         * set header font to bold style
         */
        JTableHeader tableHeader = this.getTableHeader();
        Font headerFont = tableHeader.getFont();
        tableHeader.setFont(new Font(headerFont.getName(), Font.BOLD,
                                     headerFont.getSize()));

        this.rowsWithNames = rowsWithNames;
        this.columnsWithNames = columnsWithNames;
    }

    /**
     * Overrides <code>JTable</code>'s <code>prepareRenderer</code> method.
     * Prepares the renderer by querying the data model for the
     * value and selection state
     * of the cell at <code>row</code>, <code>column</code>.
     * Returns the component (may be a <code>Component</code>
     * or a <code>JComponent</code>) under the event location.
     *
     * @param renderer  the <code>TableCellRenderer</code> to prepare
     * @param row       the row of the cell to render, where 0 is the first row
     * @param column    the column of the cell to render,
     *			where 0 is the first column
     * @return          the <code>Component</code> under the event location
     * @see javax.swing.JTable#prepareRenderer(TableCellRenderer renderer, int row, int column)
     */
    public Component prepareRenderer(TableCellRenderer renderer, int row,
                                     int column) {
        /**
         * get the component prepared from JTable's prepareRenderer method
         */
        Component c = super.prepareRenderer(renderer, row, column);

        Integer r = Integer.valueOf(row);
        Integer col = Integer.valueOf(column);

        if (rowsWithNames.contains(r)) {
            if (columnsWithNames.contains(col)) {
                c.setBackground(new Color(192, 192, 192));
            } else {
                c.setBackground(new Color(128, 255, 128));
            }
        } else {
            if (columnsWithNames.contains(col)) {
                c.setBackground(new Color(255, 128, 128));
            } else {
                c.setBackground(new Color(255, 255, 128));
            }
        }
        return c;
    }
}
