package biggests.tree;

import java.io.*;
import java.util.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.utils.*;
import smadeira.biclustering.*;

/**
 * <p>Title: Dataset Tree</p>
 *
 * <p>Description: The dynamic tree which contains the gene expression
 *                 datasets, matrices and biclusters. Allows one to add
 *                 specific nodes to it.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana Gonalves
 */
public class DatasetTree extends JTree implements Serializable {
    public static final long serialVersionUID = 3882987169020172801L;

    /**
     * The root of the datasets tree, i.e., the "Datasets" node.
     */
    private DefaultMutableTreeNode root;

    /**
     * Constructs a dataset tree with a single (mutable) tree node
     * (called "Datasets").
     */
    public DatasetTree() {
        super(new DefaultMutableTreeNode("Datasets"));
        root = (DefaultMutableTreeNode)super.getModel().getRoot();
    }

    /**
     * Adds the loaded dataset to the dataset tree and creates all additional
     * nodes for preprocessed, discretized, biclustered and post-processed
     * matrices of the dataset.<br/>
     *
     * Also shows the newly created node's path in the tree if
     * <code>showPathOfNewNode</code> is set to true.<br/>
     *
     * The node information object can be of type:<br/>
     *       <code>NodeInfo.ORIGINAL</code><br/>
     *
     *
     * @param originalNodeInfo the <code>NodeInfo</code> object of the new dataset
     * @return the <code>TreePath</code> of the new node
     * @throws InvalidNodeObjectTypeException if the type of the node object is not valid
     */
    public TreePath addLoadedDataset(NodeInfo originalNodeInfo) throws
            InvalidNodeObjectTypeException {
        /**
         * check if the given <code>NodeInfo</code> object contains an
         * original node object
         */
        int nodeObjectType = originalNodeInfo.getNodeObjectType();
        if (nodeObjectType != NodeInfo.ORIGINAL) {
            throw (new InvalidNodeObjectTypeException(
                    "Only original node information objects allowed."));
        }

        /**
         * create the new dataset
         */
        DefaultMutableTreeNode dataset =
                new DefaultMutableTreeNode(new Dataset(generateNewDatasetId()));

        /**
         * create additional nodes for the original matrix (expression matrix
         * loaded from the specified txt file), preprocessed and discretized
         * matrices and biclusters and post-processed biclusters of the dataset
         */
        dataset.add(new DefaultMutableTreeNode(originalNodeInfo));
        dataset.add(new DefaultMutableTreeNode("Preprocessed Matrices"));
        dataset.add(new DefaultMutableTreeNode("Discretized Matrices"));
        dataset.add(new DefaultMutableTreeNode("Biclusters"));
        dataset.add(new DefaultMutableTreeNode("Post-Processed Biclusters"));

        /**
         * add the newly created dataset node to the dataset tree
         */
        root.add(dataset);

        /**
         * reload the tree, so that all the new nodes get visible
         */
        ((DefaultTreeModel)this.getModel()).reload();

        Object[] path = {root, dataset, dataset.getChildAt(0)};

        return (new TreePath(path));
    }

    /**
     * Adds a new node to the datasets tree, associated with a node information object.<br/>
     *
     * Also shows the newly created node's path in the tree if
     * <code>showPathOfNewNode</code> is set to true.<br/>
     *
     * The node information object can be of type:<br/>
     *       <code>NodeInfo.PREPROCESSED</code><br/>
     *       <code>NodeInfo.DISCRETIZED</code><br/>
     *       <code>NodeInfo.BICLUSTER_GROUP</code><br/>
     *       <code>NodeInfo.POSTPROCESSED_GROUP</code>
     *
     * @param nodeInfo NodeInfo the node information object associated with the tree node
     * @param selectionPath TreePath the tree selection path
     * @return the <code>TreePath</code> of the new node
     * @throws InvalidNodeObjectTypeException when the type of the node object is not valid
     */
    public TreePath addNewNode(NodeInfo nodeInfo, TreePath selectionPath) throws
            InvalidNodeObjectTypeException {
        int nodeObjectType = nodeInfo.getNodeObjectType();
        DefaultMutableTreeNode dataset = (DefaultMutableTreeNode) selectionPath.
                                         getPath()[1];
        DefaultMutableTreeNode folder = new DefaultMutableTreeNode();

        switch (nodeObjectType) {
        case NodeInfo.PREPROCESSED:
            folder = ((DefaultMutableTreeNode) dataset.getChildAt(1));
            folder.add(new DefaultMutableTreeNode(nodeInfo));
            break;
        case NodeInfo.DISCRETIZED:
            folder = ((DefaultMutableTreeNode) dataset.getChildAt(2));
            folder.add(new DefaultMutableTreeNode(nodeInfo));
            break;
        case NodeInfo.BICLUSTER_GROUP:

            /**
             * create a new tree node for the biclusters group, with its
             * corresponding information node object
             */
            DefaultMutableTreeNode newBiclustersGroup = new
                    DefaultMutableTreeNode(nodeInfo);

            /**
             * get the set of single biclusters in the biclusters group
             */
            Bicluster[] setOfBiclusters = ((Biclustering) nodeInfo.
                                           getNodeObject()).getSetOfBiclusters();

            /**
             * add each single bicluster (creating its corresponding node
             * information object) in the biclusters group to the created
             * biclusters group tree node
             */
            for (Bicluster bic : setOfBiclusters) {
                newBiclustersGroup.add(new DefaultMutableTreeNode(
                        new NodeInfo(bic,
                                     NodeInfo.BICLUSTER_SINGLE,
                                     selectionPath,
                                     this,
                                     nodeInfo,
                                     null,
                                     null)));
            }

            /**
             * add the created biclusters group tree node to its corresponding
             * dataset in the tree
             */
            folder = ((DefaultMutableTreeNode) dataset.getChildAt(3));
            folder.add(newBiclustersGroup);
            break;
        case NodeInfo.POSTPROCESSED_GROUP:

            /**
             * create a new tree node for the post-processed biclusters group,
             * with its corresponding information node object
             */
            DefaultMutableTreeNode newPostProcessedGroup =
                    new DefaultMutableTreeNode(nodeInfo);

            /**
             * get the set of single post-processed biclusters in the
             * post-processed biclusters group
             */
            Bicluster[] setOfPostProcessedBiclusters =
                    ((Biclustering) nodeInfo.getNodeObject()).
                    getSetOfBiclusters();

            /**
             * add each single post-processed bicluster (creating its
             * corresponding node information object) in the post-processed
             * biclusters group to the created post-processed biclusters group
             * tree node
             */
            for (int j = 0; j < setOfPostProcessedBiclusters.length; j++) {
                Bicluster pos = setOfPostProcessedBiclusters[j];

                NodeInfo parentGroupNodeInfo = nodeInfo.getParentNodeInfo();

                newPostProcessedGroup.add(new DefaultMutableTreeNode(
                        new NodeInfo(pos,
                                     NodeInfo.POSTPROCESSED_SINGLE,
                                     selectionPath,
                                     this,
                                     parentGroupNodeInfo,
                                     null,
                                     null, j)));
            }

            /**
             * add the created post-processed biclusters group
             * tree node to its corresponding dataset in the tree
             */
            folder = ((DefaultMutableTreeNode) dataset.getChildAt(4));
            folder.add(newPostProcessedGroup);
            break;

        default:
            throw (new InvalidNodeObjectTypeException(
                    "Only preprocessed, discretized, bicluster group and post-processed groups allowed"));
        }
        ((DefaultTreeModel)this.getModel()).reload();

        Object[] pathAfterAddingNewNode = {root, dataset, folder,
                                          folder.getLastChild()};
        return new TreePath(pathAfterAddingNewNode);
    }

    /**
     * Adds the loaded dataset to the dataset tree and creates all additional
     * nodes for preprocessed, discretized, biclustered and post-processed
     * matrices of the dataset.<br/>
     *
     * Also shows the newly created node's path in the tree.<br/>
     *
     * The node information object can be of type:<br/>
     *       <code>NodeInfo.ORIGINAL</code><br/>
     *
     * @param originalNodeInfo the <code>NodeInfo</code> object of the new dataset
     * @param setNewPath the <code>boolean</code> which indicates if new
     *                          node's path is to display in the tree<br/>
     *                          true - display new node's path<br/>
     *                          false - do not display new node's path<br/>
     *
     * @throws InvalidNodeObjectTypeException when the type of the
     *                         <code>NodeInfo</code> is not valid
     * @see #addLoadedDataset(NodeInfo originalNodeInfo)
     */
    public void addLoadedDataset(NodeInfo originalNodeInfo, boolean setNewPath) throws
            InvalidNodeObjectTypeException {
        if (setNewPath) {
            this.setSelectionPath(addLoadedDataset(originalNodeInfo));
        } else {
            this.addLoadedDataset(originalNodeInfo);
        }
    }

    /**
     * Adds a new node to the datasets tree, associated with a node information object.
     * Also shows the newly created node's path in the tree.
     *
     * The node information object can be of type:<br/>
     *       <code>NodeInfo.PREPROCESSED</code><br/>
     *       <code>NodeInfo.DISCRETIZED</code><br/>
     *       <code>NodeInfo.BICLUSTER_GROUP</code><br/>
     *       <code>NodeInfo.POSTPROCESSED_GROUP</code>
     *
     * @param nodeInfo NodeInfo the node information object associated with the tree node
     * @param selectionPath TreePath the tree selection path
     * @param setNewPath the <code>boolean</code> which indicates if new
     *                          node's path is to display in the tree<br/>
     *                          true - display new node's path<br/>
     *                          false - do not display new node's path<br/>
     * @throws InvalidNodeObjectTypeException when the type of the node object is not valid
     * @see #addNewNode(NodeInfo nodeInfo, TreePath selectionPath)
     */
    public void addNewNode(NodeInfo nodeInfo, TreePath selectionPath,
                           boolean setNewPath) throws
            InvalidNodeObjectTypeException {
        if (setNewPath) {
            this.setSelectionPath(addNewNode(nodeInfo, selectionPath));
        } else {
            this.addNewNode(nodeInfo, selectionPath);
        }
    }

    /**
     * Reloads the tree model, so all (new) nodes get visible.
     */
    public void reload() {
        ((DefaultTreeModel)this.getModel()).reload();
    }

    /**
     * Returns the root node of the dataset tree.
     *
     * @return DefaultMutableTreeNode the root node of the dataset
     */
    public DefaultMutableTreeNode getRoot() {
        return root;
    }

    /**
     * Returns the next id for a new <code>NodeInfo</code> object to be
     * created in the dataset of the given selected path. The next id depends on:
     * <br/>
     *     - the dataset in which the node information object will be included
     *     (accessible through the selected path)
     * <br/>
     *     - the data type of the node information object to be created
     *
     * @param nodeObjectType int the type of the node object (so we know in which
     * group of node information objects we have to look for the next id):<br/>
     *    <code>NodeInfo.PREPROCESSED</code><br/>
     *    <code>NodeInfo.DISCRETIZED</code><br/>
     *    <code>NodeInfo.BICLUSTER_GROUP</code><br/>
     *    <code>NodeInfo.POSTPROCESSED_GROUP</code><br/>
     * @param selectionPath TreePath the selected path
     * @param update <code>boolean</code> <code>true</code> to update the
     *               number of nodes created in the tree for this
     *               <code>nodeObjectType</code> so far; <code>false</code>
     *               if one only wants to know which id is next, without
     *               inserting a new node in the tree
     * @return int the new id for a new <code>NodeInfo<code/> object
     * @throws InvalidNodeObjectTypeException if the type of the node object is not valid
     */
    public int generateNewNodeInfoId(int nodeObjectType, TreePath selectionPath,
                                     boolean update) throws
            InvalidNodeObjectTypeException {
        if (nodeObjectType != NodeInfo.PREPROCESSED &&
            nodeObjectType != NodeInfo.DISCRETIZED &&
            nodeObjectType != NodeInfo.BICLUSTER_GROUP &&
            nodeObjectType != NodeInfo.POSTPROCESSED_GROUP) {
            throw (new InvalidNodeObjectTypeException(
                    "Only preprocessed, discretized, bicluster groups and post-processed groups are allowed"));
        }

        /**
         * get all the path nodes of the given selectionPath
         */
        Object[] path = selectionPath.getPath();

        /**
         * get the dataset of the given selection path
         */
        Dataset d = (Dataset) ((DefaultMutableTreeNode) path[1]).getUserObject();

        /**
         * order (from the given dataset) and return a new id for the given
         * node object type
         */
        return (d.generateNewNodeObjectTypeId(nodeObjectType, update));
    }

    /**
     * Generates a dataset id for a new <code>Dataset</code> object.
     *
     * @return int the new dataset id
     */
    public int generateNewDatasetId() {
        int maxId = -1, id;

        /**
         * get the set of all datasets in the tree
         */
        Enumeration e = root.children();

        /**
         * find the maximum id value for the datasets set
         */
        while (e.hasMoreElements()) {
            id = ((Dataset) ((DefaultMutableTreeNode) e.nextElement()).
                  getUserObject()).getId();
            if (id > maxId) {
                maxId = id;
            }
        }

        /**
         * return the new id, which is the maximum id found plus 1
         */
        return (maxId + 1);
    }

    /**
     * Looks for a <code>NodeInfo</code> object of a bicluster or
     * post-processed bicluster object with a specific <code>id</code>.
     * Returns the corresponding <code>NodeInfo</code> object if search
     * succeeds and null otherwise.
     *
     * @param setOfBiclustersNodes <code>Enumeration</code> with the set of
     * bicluster nodes where to look for the bicluster node information
     * object with the given <code>id</code>; each element of the
     * <code>setOfBiclustersNodes</code> has type
     * <code>DefaultMutableTreeNode</code> and contains a <code>NodeInfo</code>
     * object which, in turn, contains a bicluster or post-processed
     * bicluster object
     * @param id the <code>int</code> of the bicluster to search for
     * @return NodeInfo the node information object which contains the
     * bicluster with the given <code>id</code>; null if bicluster was not found
     */
    public NodeInfo findBiclusterNodeByIDinSetOfBiclustersNodes(
            Enumeration setOfBiclustersNodes,
            int id) {
        while (setOfBiclustersNodes.hasMoreElements()) {
            NodeInfo biclusterNodeInfo = (NodeInfo) ((
                    DefaultMutableTreeNode)
                    setOfBiclustersNodes.nextElement()).
                                         getUserObject();
            if (biclusterNodeInfo.getId() == id) {
                return biclusterNodeInfo;
            }
        }
        return null;
    }
}
